"""
Construction helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class ConstructionHelper:
  """
  Construction helper
  """
  # NREL

  _reference_city_to_nrel_climate_zone = {
    'Miami': 'ASHRAE_2004:1A',
    'Houston': 'ASHRAE_2004:2A',
    'Phoenix': 'ASHRAE_2004:2B',
    'Atlanta': 'ASHRAE_2004:3A',
    'Los Angeles': 'ASHRAE_2004:3B',
    'Las Vegas': 'ASHRAE_2004:3B',
    'San Francisco': 'ASHRAE_2004:3C',
    'Baltimore': 'ASHRAE_2004:4A',
    'Albuquerque': 'ASHRAE_2004:4B',
    'Seattle': 'ASHRAE_2004:4C',
    'Chicago': 'ASHRAE_2004:5A',
    'Boulder': 'ASHRAE_2004:5B',
    'Minneapolis': 'ASHRAE_2004:6A',
    'Helena': 'ASHRAE_2004:6B',
    'Duluth': 'ASHRAE_2004:7A',
    'Fairbanks': 'ASHRAE_2004:8A'
  }

  _reference_city_to_nrcan_climate_zone = {
    'Montreal': '6',
    'Repentigny': '6',
    "Montreal Int'l": '6',
    'Levis': '7A',
    'Quebec City': '7A',
    'Kelowna': '5',
    'Park Slope': '4',
    'Varennes': '6',
    'Laval': '6',
    'Longueuil': '6',
    'Mont-Royal': '6',
    'Deux-Montagnes': '6',
    'Dorval': '6',
    'Dollard-Des Ormeaux': '6',
    'Notre-Dame-de-Grace': '6',
    'Westmount': '6',
    'Cote-Saint-Luc': '6',
    'Kirkland': '6',
    'Beaconsfield': '6',
    'Sainte-Catherine': '6',
    'Vaudreuil-Dorion': '6',
    'Saint-Laurent': '6',
    'Pointe-Claire': '6',
    'Boucherville': '6',
    'Mascouche': '6',
    'Saint-Leonard': '6',
    'La Prairie': '6'
  }

  _reference_city_to_israel_climate_zone = {
    'Eilat': 'BWh'
  }

  _reference_city_to_palma_climate_zone ={
    'Palma': 'B3'
  }

  @staticmethod
  def yoc_to_nrel_standard(year_of_construction):
    """
    Year of construction to NREL standard
    :param year_of_construction: int
    :return: str
    """
    if int(year_of_construction) < 2009:
      standard = 'ASHRAE 90.1_2004'
    else:
      standard = 'ASHRAE 189.1_2009'
    return standard

  @staticmethod
  def city_to_nrel_climate_zone(reference_city):
    """
    City name to NREL climate zone
    :param reference_city: str
    :return: str
    """
    # todo: finish dictionary implementation
    if reference_city not in ConstructionHelper._reference_city_to_nrel_climate_zone:
      reference_city = 'Baltimore'
    return ConstructionHelper._reference_city_to_nrel_climate_zone[reference_city]

  @staticmethod
  def city_to_nrcan_climate_zone(reference_city):
    """
    City name to NRCAN climate zone
    :param reference_city: str
    :return: str
    """
    return ConstructionHelper._reference_city_to_nrcan_climate_zone[reference_city]

  @staticmethod
  def city_to_israel_climate_zone(reference_city):
    """
    City name to Israel climate zone
    :param reference_city: str
    :return: str
    """
    return ConstructionHelper._reference_city_to_israel_climate_zone[reference_city]

  @staticmethod
  def city_to_palma_climate_zone(reference_city):
    """
    City name to Palma climate zone
    :param reference_city: str
    :return: str
    """
    return ConstructionHelper._reference_city_to_palma_climate_zone[reference_city]




"""
Energy storage system. Abstract class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from abc import ABC


class EnergyStorageSystem(ABC):
  """
  Energy storage System class
  """
  def __init__(self):
    self._type_energy_stored = None
    self._storage_type = None
    self._model_name = None
    self._manufacturer = None
    self._nominal_capacity = None
    self._losses_ratio = None

  @property
  def type_energy_stored(self):
    """
    Get type of energy stored from ['electrical', 'thermal']
    :return: string
    """
    return self._type_energy_stored

  @type_energy_stored.setter
  def type_energy_stored(self, value):
    """
    Set type of energy stored from ['electrical', 'thermal']
    :return: string
    """
    self._type_energy_stored = value

  @property
  def storage_type(self):
    """
    Get storage type
    :return: string
    """
    return self._storage_type

  @storage_type.setter
  def storage_type(self, value):
    """
    Get storage type
    :param value: string
    """
    self._storage_type = value

  @property
  def model_name(self):
    """
    Get system model
    :return: string
    """
    return self._model_name

  @model_name.setter
  def model_name(self, value):
    """
    Set system model
    :param value: string
    """
    self._model_name = value

  @property
  def manufacturer(self):
    """
    Get name of manufacturer
    :return: string
    """
    return self._manufacturer

  @manufacturer.setter
  def manufacturer(self, value):
    """
    Set name of manufacturer
    :param value: string
    """
    self._manufacturer = value

  @property
  def nominal_capacity(self):
    """
    Get the nominal capacity of storage systems in Jules
    :return: float
    """
    return self._nominal_capacity

  @nominal_capacity.setter
  def nominal_capacity(self, value):
    """
    Set the nominal capacity of storage systems in Jules
    :return: float
    """
    self._nominal_capacity = value

  @property
  def losses_ratio(self):
    """
    Get the losses-ratio of storage system in Jules lost / Jules stored
    :return: float
    """
    return self._losses_ratio

  @losses_ratio.setter
  def losses_ratio(self, value):
    """
    Set the losses-ratio of storage system in Jules lost / Jules stored
    :return: float
    """
    self._losses_ratio = value



"""
Construction helper module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""
from hub.helpers import constants as cte


class ConstructionHelper:
  """
  Construction helper class
  """
  _reference_standard_to_construction_period = {
    'non_standard_dompark': '1900 - 2004',
    'ASHRAE 90.1_2004': '2004 - 2009',
    'ASHRAE 189.1_2009': '2009 - PRESENT'
  }

  _nrel_surfaces_types_to_hub_types = {
    'exterior wall': cte.WALL,
    'interior wall': cte.INTERIOR_WALL,
    'ground wall': cte.GROUND_WALL,
    'exterior slab': cte.GROUND,
    'attic floor': cte.ATTIC_FLOOR,
    'interior slab': cte.INTERIOR_SLAB,
    'roof': cte.ROOF
  }

  _nrcan_surfaces_types_to_hub_types = {
    'OutdoorsWall': cte.WALL,
    'OutdoorsRoofCeiling': cte.ROOF,
    'OutdoorsFloor': cte.ATTIC_FLOOR,
    'Window': cte.WINDOW,
    'Skylight': cte.SKYLIGHT,
    'GroundWall': cte.GROUND_WALL,
    'GroundRoofCeiling': cte.GROUND_WALL,
    'GroundFloor': cte.GROUND
  }

  @property
  def reference_standard_to_construction_period(self):
    """
    Get reference standard to construction period dictionary
    :return: {}
    """
    return self._reference_standard_to_construction_period

  @property
  def nrel_surfaces_types_to_hub_types(self):
    """
    Get reference nrel surface type to hub type dictionary
    :return: {}
    """
    return self._nrel_surfaces_types_to_hub_types

  @property
  def nrcan_surfaces_types_to_hub_types(self):
    """
    Get reference nrcan surface type to hub type dictionary
    :return: {}
    """
    return self._nrcan_surfaces_types_to_hub_types



"""
Energy System catalog archetype, understood as a cluster of energy systems
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from typing import List

from hub.catalog_factories.data_models.energy_systems.system import System


class Archetype:
  """
  Archetype class
  """

  def __init__(self, name, systems, archetype_cluster_id=None):
    self._cluster_id = archetype_cluster_id
    self._name = name
    self._systems = systems

  @property
  def cluster_id(self):
    """
    Get id
    :return: string
    """
    return self._cluster_id

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return self._name

  @property
  def systems(self) -> List[System]:
    """
    Get list of equipments that compose the total energy system
    :return: [Equipment]
    """
    return self._systems

  def to_dictionary(self):
    """Class content to dictionary"""
    _systems = []
    for _system in self.systems:
      _systems.append(_system.to_dictionary())
    content = {
      'Archetype': {
        'cluster_id': self.cluster_id,
        'name': self.name,
        'systems': _systems
      }
    }
    return content



"""
Application repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime
import logging

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.session import Session

from hub.persistence.repository import Repository
from hub.persistence.models import Application as Model


class Application(Repository):
  """
  Application repository
  """
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(Application, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, description: str, application_uuid: str):
    """
    Inserts a new application
    :param name: Application name
    :param description: Application description
    :param application_uuid: Unique identifier for the application
    :return: Identity id
    """
    try:
      application = self.get_by_uuid(application_uuid)
      if application is not None:
        raise SQLAlchemyError('application already exists')
    except TypeError:
      pass
    try:
      application = Model(name=name, description=description, application_uuid=application_uuid)
      with Session(self.engine) as session:
        session.add(application)
        session.commit()
        session.refresh(application)
        return application.id
    except SQLAlchemyError as err:
      logging.error('An error occurred while creating application %s', err)
      raise SQLAlchemyError from err

  def update(self, application_uuid: str, name: str, description: str):
    """
    Updates an application
    :param application_uuid: the application uuid of the application to be updated
    :param name: the application name
    :param description: the application description
    :return: None
    """
    try:
      with Session(self.engine) as session:
        session.query(Model).filter(
          Model.application_uuid == application_uuid
        ).update({'name': name, 'description': description, 'updated': datetime.datetime.utcnow()})
        session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while updating application %s', err)
      raise SQLAlchemyError from err

  def delete(self, application_uuid: str):
    """
    Deletes an application with the application_uuid
    :param application_uuid: The application uuid
    :return: None
    """
    try:
      with Session(self.engine) as session:
        session.query(Model).filter(Model.application_uuid == application_uuid).delete()
        session.flush()
        session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while deleting application %s', err)
      raise SQLAlchemyError from err

  def get_by_uuid(self, application_uuid: str) -> Model:
    """
    Fetch Application based on the application uuid
    :param application_uuid: the application uuid
    :return: Application with the provided application_uuid
    """
    try:
      with Session(self.engine) as session:
        result_set = session.execute(select(Model).where(
          Model.application_uuid == application_uuid)
        ).first()
        return result_set[0]
    except SQLAlchemyError as err:
      logging.error('Error while fetching application by application_uuid %s', err)
      raise SQLAlchemyError from err
    except TypeError as err:
      logging.error('Error while fetching application, empty result %s', err)
      raise TypeError from err



"""
Energy Systems catalog System
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from typing import Union, List
from pathlib import Path

from hub.catalog_factories.data_models.energy_systems.generation_system import GenerationSystem
from hub.catalog_factories.data_models.energy_systems.distribution_system import DistributionSystem


class System:
  """
  System class
  """

  def __init__(self,
               system_id,
               demand_types,
               name=None,
               generation_systems=None,
               distribution_systems=None,
               configuration_schema=None):
    self._system_id = system_id
    self._name = name
    self._demand_types = demand_types
    self._generation_systems = generation_systems
    self._distribution_systems = distribution_systems
    self._configuration_schema = configuration_schema

  @property
  def id(self):
    """
    Get equipment id
    :return: string
    """
    return self._system_id

  @property
  def name(self):
    """
    Get the system name
    :return: string
    """
    return self._name

  @property
  def demand_types(self):
    """
    Get demand able to cover from ['heating', 'cooling', 'domestic_hot_water', 'electricity']
    :return: [string]
    """
    return self._demand_types

  @property
  def generation_systems(self) -> Union[None, List[GenerationSystem]]:
    """
    Get generation systems
    :return: [GenerationSystem]
    """
    return self._generation_systems

  @property
  def distribution_systems(self) -> Union[None, List[DistributionSystem]]:
    """
    Get distribution systems
    :return: [DistributionSystem]
    """
    return self._distribution_systems

  @property
  def configuration_schema(self) -> Path:
    """
    Get system configuration schema
    :return: Path
    """
    return self._configuration_schema

  def to_dictionary(self):
    """Class content to dictionary"""
    _generation_systems = []
    for _generation in self.generation_systems:
      _generation_systems.append(_generation.to_dictionary())
    _distribution_systems = [_distribution.to_dictionary() for _distribution in
                             self.distribution_systems] if self.distribution_systems is not None else None

    content = {'system': {'id': self.id,
                          'name': self.name,
                          'demand types': self.demand_types,
                          'generation system(s)': _generation_systems,
                          'distribution system(s)': _distribution_systems,
                          'configuration schema path': self.configuration_schema
                          }
               }
    return content



"""
TestSystemsFactory
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import subprocess
from pathlib import Path
from unittest import TestCase
import copy
from typing import cast

import hub.helpers.constants as cte
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.exports.exports_factory import ExportsFactory
from hub.helpers.dictionaries import Dictionaries
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.weather_factory import WeatherFactory
from hub.imports.results_factory import ResultFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.city_model_structure.energy_systems.generation_system import GenerationSystem
from hub.city_model_structure.energy_systems.distribution_system import DistributionSystem
from hub.city_model_structure.energy_systems.emission_system import EmissionSystem
from hub.city_model_structure.energy_systems.thermal_storage_system import ThermalStorageSystem
from hub.city_model_structure.energy_systems.non_pv_generation_system import NonPvGenerationSystem


class TestSystemsFactory(TestCase):
  """
  TestSystemsFactory TestCase
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()
    file = 'test.geojson'
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('geojson',
                                 path=file_path,
                                 height_field='citygml_me',
                                 year_of_construction_field='ANNEE_CONS',
                                 function_field='CODE_UTILI',
                                 function_to_hub=Dictionaries().montreal_function_to_hub_function).city

  def test_montreal_custom_system_factory(self):
    """
    Enrich the city with the construction information and verify it
    """
    for building in self._city.buildings:
      building.energy_systems_archetype_name = 'system 1 gas'

    EnergySystemsFactory('montreal_custom', self._city).enrich()
    self.assertEqual(1, len(self._city.generic_energy_systems))

  def test_montreal_custom_system_results(self):
    """
    Enrich the city with the construction information and verify it
    """
    ConstructionFactory('nrcan', self._city).enrich()
    UsageFactory('nrcan', self._city).enrich()
    WeatherFactory('epw', self._city).enrich()
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    EnergyBuildingsExportsFactory('insel_monthly_energy_balance', self._city, self._output_path).export()
    for building in self._city.buildings:
      insel_path = (self._output_path / f'{building.name}.insel')
      subprocess.run(['insel', str(insel_path)])
    ResultFactory('insel_monthly_energy_balance', self._city, self._output_path).enrich()

    for building in self._city.buildings:
      building.energy_systems_archetype_name = 'system 1 gas pv'
    EnergySystemsFactory('montreal_custom', self._city).enrich()
    # Need to assign energy systems to buildings:
    for building in self._city.buildings:
      _building_energy_systems = []
      for energy_system in building.energy_systems:
        if cte.HEATING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.heat_power = building.heating_peak_load[cte.YEAR][0]
        if cte.COOLING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.cooling_power = building.cooling_peak_load[cte.YEAR][0]

    for building in self._city.buildings:
      self.assertLess(0, building.heating_consumption[cte.YEAR][0])
      self.assertLess(0, building.cooling_consumption[cte.YEAR][0])
      self.assertLess(0, building.domestic_hot_water_consumption[cte.YEAR][0])
      self.assertLess(0, building.onsite_electrical_production[cte.YEAR][0])

  def test_montreal_future_system_results(self):
    """
    Enrich the city with the construction information and verify it
    """
    ConstructionFactory('nrcan', self._city).enrich()
    UsageFactory('nrcan', self._city).enrich()
    WeatherFactory('epw', self._city).enrich()
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    EnergyBuildingsExportsFactory('insel_monthly_energy_balance', self._city, self._output_path).export()
    for building in self._city.buildings:
      insel_path = (self._output_path / f'{building.name}.insel')
      subprocess.run(['insel', str(insel_path)])
    ResultFactory('insel_monthly_energy_balance', self._city, self._output_path).enrich()

    for building in self._city.buildings:
      building.energy_systems_archetype_name = ('Central Hydronic Air and Gas Source Heating System with Unitary Split '
                                                'Cooling and Air Source HP DHW and Grid Tied PV')
    EnergySystemsFactory('montreal_future', self._city).enrich()
    # Need to assign energy systems to buildings:
    for building in self._city.buildings:
      _building_energy_systems = []
      for energy_system in building.energy_systems:
        if cte.HEATING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.nominal_heat_output = building.heating_peak_load[cte.YEAR][0]
        if cte.COOLING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.nominal_cooling_output = building.cooling_peak_load[cte.YEAR][0]

    for building in self._city.buildings:
      self.assertLess(0, building.heating_consumption[cte.YEAR][0])
      self.assertLess(0, building.cooling_consumption[cte.YEAR][0])
      self.assertLess(0, building.domestic_hot_water_consumption[cte.YEAR][0])
      if 'PV' in building.energy_systems_archetype_name:
        self.assertLess(0, building.onsite_electrical_production[cte.YEAR][0])

  def test_palma_system_results(self):
    """
    Enrich the city with the construction information and verify it
    """
    ConstructionFactory('nrcan', self._city).enrich()
    UsageFactory('nrcan', self._city).enrich()
    WeatherFactory('epw', self._city).enrich()
    ExportsFactory('sra', self._city, self._output_path).export()
    sra_path = (self._output_path / f'{self._city.name}_sra.xml').resolve()
    subprocess.run(['sra', str(sra_path)])
    ResultFactory('sra', self._city, self._output_path).enrich()
    EnergyBuildingsExportsFactory('insel_monthly_energy_balance', self._city, self._output_path).export()
    for building in self._city.buildings:
      insel_path = (self._output_path / f'{building.name}.insel')
      subprocess.run(['insel', str(insel_path)])
    ResultFactory('insel_monthly_energy_balance', self._city, self._output_path).enrich()

    for building in self._city.buildings:
      building.energy_systems_archetype_name = 'PV and heat pump'
    EnergySystemsFactory('palma', self._city).enrich()
    # Need to assign energy systems to buildings:
    for building in self._city.buildings:
      _building_energy_systems = []
      for energy_system in building.energy_systems:
        if cte.HEATING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.nominal_heat_output = building.heating_peak_load[cte.YEAR][0]
        if cte.COOLING in energy_system.demand_types:
          _generation_system = cast(NonPvGenerationSystem, energy_system.generation_systems[0])
          _generation_system.nominal_cooling_output = building.cooling_peak_load[cte.YEAR][0]

    for building in self._city.buildings:
      self.assertLess(0, building.heating_consumption[cte.YEAR][0])
      self.assertLess(0, building.cooling_consumption[cte.YEAR][0])
      self.assertLess(0, building.domestic_hot_water_consumption[cte.YEAR][0])
      self.assertLess(0, building.onsite_electrical_production[cte.YEAR][0])


"""
Insel monthly energy balance
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Project collaborator Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from hub.helpers.monthly_values import MonthlyValues
import csv

import hub.helpers.constants as cte


class EnergyPlusMultipleBuildings:
  def __init__(self, city, base_path):
    self._city = city
    self._base_path = base_path

  def _building_energy_demands(self, energy_plus_output_file_path):
    buildings_energy_demands = {}

    with open(Path(energy_plus_output_file_path).resolve(), 'r', encoding='utf8') as csv_file:
      csv_output = list(csv.DictReader(csv_file))

      for building in self._city.buildings:
        building_name = building.name.upper()
        buildings_energy_demands[f'Building {building_name} Heating Demand (J)'] = [
          float(
            row[f"{building_name} IDEAL LOADS AIR SYSTEM:Zone Ideal Loads Supply Air Total Heating Energy [J](Hourly)"])
          for row in csv_output
        ]
        buildings_energy_demands[f'Building {building_name} Cooling Demand (J)'] = [
          float(
            row[f"{building_name} IDEAL LOADS AIR SYSTEM:Zone Ideal Loads Supply Air Total Cooling Energy [J](Hourly)"])
          for row in csv_output
        ]
        buildings_energy_demands[f'Building {building_name} DHW Demand (W)'] = [
          float(row[f"DHW {building_name}:Water Use Equipment Heating Rate [W](Hourly)"])
          for row in csv_output
        ]
        buildings_energy_demands[f'Building {building_name} Appliances (W)'] = [
          float(row[f"{building_name}_APPLIANCE:Other Equipment Electricity Rate [W](Hourly)"])
          for row in csv_output
        ]
        buildings_energy_demands[f'Building {building_name} Lighting (W)'] = [
          float(row[f"{building_name}:Zone Lights Electricity Rate [W](Hourly)"]) for row in csv_output
        ]
    return buildings_energy_demands

  def enrich(self):
    """
    Enrich the city by using the energy plus workflow output files (J)
    :return: None
    """
    file_name = f'{self._city.name}_out.csv'
    energy_plus_output_file_path = Path(self._base_path / file_name).resolve()
    if energy_plus_output_file_path.is_file():
      building_energy_demands = self._building_energy_demands(energy_plus_output_file_path)
      for building in self._city.buildings:
        building_name = building.name.upper()
        building.heating_demand[cte.HOUR] = building_energy_demands[f'Building {building_name} Heating Demand (J)']
        building.cooling_demand[cte.HOUR] = building_energy_demands[f'Building {building_name} Cooling Demand (J)']
        building.domestic_hot_water_heat_demand[cte.HOUR] = \
            [x * cte.WATTS_HOUR_TO_JULES for x in building_energy_demands[f'Building {building_name} DHW Demand (W)']]
        building.appliances_electrical_demand[cte.HOUR] = \
            [x * cte.WATTS_HOUR_TO_JULES for x in building_energy_demands[f'Building {building_name} Appliances (W)']]
        building.lighting_electrical_demand[cte.HOUR] = \
            [x * cte.WATTS_HOUR_TO_JULES for x in building_energy_demands[f'Building {building_name} Lighting (W)']]
        building.heating_demand[cte.MONTH] = MonthlyValues.get_total_month(building.heating_demand[cte.HOUR])
        building.cooling_demand[cte.MONTH] = MonthlyValues.get_total_month(building.cooling_demand[cte.HOUR])
        building.domestic_hot_water_heat_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.domestic_hot_water_heat_demand[cte.HOUR]))
        building.appliances_electrical_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.appliances_electrical_demand[cte.HOUR]))
        building.lighting_electrical_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.lighting_electrical_demand[cte.HOUR]))
        building.heating_demand[cte.YEAR] = [sum(building.heating_demand[cte.MONTH])]
        building.cooling_demand[cte.YEAR] = [sum(building.cooling_demand[cte.MONTH])]
        building.domestic_hot_water_heat_demand[cte.YEAR] = [sum(building.domestic_hot_water_heat_demand[cte.MONTH])]
        building.appliances_electrical_demand[cte.YEAR] = [sum(building.appliances_electrical_demand[cte.MONTH])]
        building.lighting_electrical_demand[cte.YEAR] = [sum(building.lighting_electrical_demand[cte.MONTH])]




"""
Montreal future system importer
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca

"""

import logging
import copy

from hub.catalog_factories.energy_systems_catalog_factory import EnergySystemsCatalogFactory
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.city_model_structure.energy_systems.distribution_system import DistributionSystem
from hub.city_model_structure.energy_systems.non_pv_generation_system import NonPvGenerationSystem
from hub.city_model_structure.energy_systems.pv_generation_system import PvGenerationSystem
from hub.city_model_structure.energy_systems.electrical_storage_system import ElectricalStorageSystem
from hub.city_model_structure.energy_systems.thermal_storage_system import ThermalStorageSystem
from hub.city_model_structure.energy_systems.emission_system import EmissionSystem
from hub.helpers.dictionaries import Dictionaries


class MontrealFutureEnergySystemParameters:
  """
  MontrealCustomEnergySystemParameters class
  """

  def __init__(self, city):
    self._city = city

  def enrich_buildings(self):
    """
    Returns the city with the system parameters assigned to the buildings
    :return:
    """
    city = self._city
    montreal_custom_catalog = EnergySystemsCatalogFactory('montreal_future').catalog
    if city.generic_energy_systems is None:
      _generic_energy_systems = {}
    else:
      _generic_energy_systems = city.generic_energy_systems
    for building in city.buildings:
      archetype_name = building.energy_systems_archetype_name
      try:
        archetype = self._search_archetypes(montreal_custom_catalog, archetype_name)
        building.energy_systems_archetype_cluster_id = archetype.cluster_id
      except KeyError:
        logging.error('Building %s has unknown energy system archetype for system name %s', building.name,
                      archetype_name)
        continue

      if archetype.name not in _generic_energy_systems:
        _generic_energy_systems = self._create_generic_systems_list(archetype, _generic_energy_systems)

    city.generic_energy_systems = _generic_energy_systems

    self._assign_energy_systems_to_buildings(city)

  @staticmethod
  def _search_archetypes(catalog, name):
    archetypes = catalog.entries('archetypes')
    for building_archetype in archetypes:
      if str(name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  def _create_generic_systems_list(self, archetype, _generic_energy_systems):
    building_systems = []
    for archetype_system in archetype.systems:
      energy_system = EnergySystem()
      _hub_demand_types = []
      for demand_type in archetype_system.demand_types:
        _hub_demand_types.append(Dictionaries().montreal_demand_type_to_hub_energy_demand_type[demand_type])
      energy_system.name = archetype_system.name
      energy_system.demand_types = _hub_demand_types
      energy_system.configuration_schema = archetype_system.configuration_schema
      energy_system.generation_systems = self._create_generation_systems(archetype_system)
      if energy_system.distribution_systems is not None:
        energy_system.distribution_systems = self._create_distribution_systems(archetype_system)
      building_systems.append(energy_system)

    _generic_energy_systems[archetype.name] = building_systems

    return _generic_energy_systems

  def _create_generation_systems(self, archetype_system):
    _generation_systems = []
    archetype_generation_systems = archetype_system.generation_systems
    if archetype_generation_systems is not None:
      for archetype_generation_system in archetype_system.generation_systems:
        if archetype_generation_system.system_type == 'photovoltaic':
          _generation_system = PvGenerationSystem()
          _generation_system.name = archetype_generation_system.name
          _generation_system.model_name = archetype_generation_system.model_name
          _generation_system.manufacturer = archetype_generation_system.manufacturer
          _type = archetype_generation_system.system_type
          _generation_system.system_type = Dictionaries().montreal_generation_system_to_hub_energy_generation_system[_type]
          _fuel_type = Dictionaries().montreal_custom_fuel_to_hub_fuel[archetype_generation_system.fuel_type]
          _generation_system.fuel_type = _fuel_type
          _generation_system.electricity_efficiency = archetype_generation_system.electricity_efficiency
          _generation_system.nominal_electricity_output = archetype_generation_system.nominal_electricity_output
          _generation_system.nominal_ambient_temperature = archetype_generation_system.nominal_ambient_temperature
          _generation_system.nominal_cell_temperature = archetype_generation_system.nominal_cell_temperature
          _generation_system.nominal_radiation = archetype_generation_system.nominal_radiation
          _generation_system.standard_test_condition_cell_temperature = archetype_generation_system.standard_test_condition_cell_temperature
          _generation_system.standard_test_condition_maximum_power = archetype_generation_system.standard_test_condition_maximum_power
          _generation_system.standard_test_condition_radiation = archetype_generation_system.standard_test_condition_radiation
          _generation_system.cell_temperature_coefficient = archetype_generation_system.cell_temperature_coefficient
          _generation_system.width = archetype_generation_system.width
          _generation_system.height = archetype_generation_system.height
          _generation_system.tilt_angle = self._city.latitude
          _generic_storage_system = None
          if archetype_generation_system.energy_storage_systems is not None:
            _storage_systems = []
            for storage_system in archetype_generation_system.energy_storage_systems:
              if storage_system.type_energy_stored == 'electrical':
                _generic_storage_system = ElectricalStorageSystem()
                _generic_storage_system.type_energy_stored = 'electrical'
              _storage_systems.append(_generic_storage_system)
            _generation_system.energy_storage_systems = _storage_systems

        else:
          _generation_system = NonPvGenerationSystem()
          _generation_system.name = archetype_generation_system.name
          _generation_system.model_name = archetype_generation_system.model_name
          _generation_system.manufacturer = archetype_generation_system.manufacturer
          _type = archetype_generation_system.system_type
          _generation_system.system_type = Dictionaries().montreal_generation_system_to_hub_energy_generation_system[_type]
          _fuel_type = Dictionaries().montreal_custom_fuel_to_hub_fuel[archetype_generation_system.fuel_type]
          _generation_system.fuel_type = _fuel_type
          _generation_system.nominal_heat_output = archetype_generation_system.nominal_heat_output
          _generation_system.nominal_cooling_output = archetype_generation_system.nominal_cooling_output
          _generation_system.maximum_heat_output = archetype_generation_system.maximum_heat_output
          _generation_system.minimum_heat_output = archetype_generation_system.minimum_heat_output
          _generation_system.maximum_cooling_output = archetype_generation_system.maximum_cooling_output
          _generation_system.minimum_cooling_output = archetype_generation_system.minimum_cooling_output
          _generation_system.source_temperature = archetype_generation_system.source_temperature
          _generation_system.source_mass_flow = archetype_generation_system.source_mass_flow
          _generation_system.supply_medium = archetype_generation_system.supply_medium
          _generation_system.maximum_heat_supply_temperature = archetype_generation_system.maximum_heat_supply_temperature
          _generation_system.maximum_cooling_supply_temperature = archetype_generation_system.maximum_cooling_supply_temperature
          _generation_system.minimum_heat_supply_temperature = archetype_generation_system.minimum_heat_supply_temperature
          _generation_system.minimum_cooling_supply_temperature = archetype_generation_system.minimum_cooling_supply_temperature
          _generation_system.heat_output_curve = archetype_generation_system.heat_output_curve
          _generation_system.heat_fuel_consumption_curve = archetype_generation_system.heat_fuel_consumption_curve
          _generation_system.heat_efficiency_curve = archetype_generation_system.heat_efficiency_curve
          _generation_system.cooling_output_curve = archetype_generation_system.cooling_output_curve
          _generation_system.cooling_fuel_consumption_curve = archetype_generation_system.cooling_fuel_consumption_curve
          _generation_system.cooling_efficiency_curve = archetype_generation_system.cooling_efficiency_curve
          _generation_system.domestic_hot_water = archetype_generation_system.domestic_hot_water
          _generation_system.nominal_electricity_output = archetype_generation_system.nominal_electricity_output
          _generation_system.source_medium = archetype_generation_system.source_medium
          _generation_system.heat_efficiency = archetype_generation_system.heat_efficiency
          _generation_system.cooling_efficiency = archetype_generation_system.cooling_efficiency
          _generation_system.electricity_efficiency = archetype_generation_system.electricity_efficiency
          _generation_system.reversibility = archetype_generation_system.reversibility
          _generic_storage_system = None
          if archetype_generation_system.energy_storage_systems is not None:
            _storage_systems = []
            for storage_system in archetype_generation_system.energy_storage_systems:
              if storage_system.type_energy_stored == 'electrical':
                _generic_storage_system = ElectricalStorageSystem()
                _generic_storage_system.type_energy_stored = 'electrical'
              else:
                _generic_storage_system = ThermalStorageSystem()
                _generic_storage_system.type_energy_stored = storage_system.type_energy_stored
                _generic_storage_system.height = storage_system.height
                _generic_storage_system.layers = storage_system.layers
                _generic_storage_system.storage_medium = storage_system.storage_medium
                _generic_storage_system.heating_coil_capacity = storage_system.heating_coil_capacity
              _storage_systems.append(_generic_storage_system)
            _generation_system.energy_storage_systems = _storage_systems
          if archetype_generation_system.domestic_hot_water:
            _generation_system.domestic_hot_water = True
          if archetype_generation_system.reversibility:
            _generation_system.reversibility = True
          if archetype_generation_system.simultaneous_heat_cold:
            _generation_system.simultaneous_heat_cold = True
        _generation_systems.append(_generation_system)
    return _generation_systems

  @staticmethod
  def _create_distribution_systems(archetype_system):
    _distribution_systems = []
    archetype_distribution_systems = archetype_system.distribution_systems
    if archetype_distribution_systems is not None:
      for archetype_distribution_system in archetype_system.distribution_systems:
        _distribution_system = DistributionSystem()
        _distribution_system.type = archetype_distribution_system.type
        _distribution_system.distribution_consumption_fix_flow = \
          archetype_distribution_system.distribution_consumption_fix_flow
        _distribution_system.distribution_consumption_variable_flow = \
          archetype_distribution_system.distribution_consumption_variable_flow
        _distribution_system.heat_losses = archetype_distribution_system.heat_losses
        _generic_emission_system = None
        if archetype_distribution_system.emission_systems is not None:
          _emission_systems = []
          for emission_system in archetype_distribution_system.emission_systems:
            _generic_emission_system = EmissionSystem()
            _generic_emission_system.parasitic_energy_consumption = emission_system.parasitic_energy_consumption
            _emission_systems.append(_generic_emission_system)
          _distribution_system.emission_systems = _emission_systems
        _distribution_systems.append(_distribution_system)
      return _distribution_systems

  @staticmethod
  def _assign_energy_systems_to_buildings(city):
    for building in city.buildings:
      _building_energy_systems = []
      energy_systems_cluster_name = building.energy_systems_archetype_name
      if str(energy_systems_cluster_name) == 'nan':
        break
      _generic_building_energy_systems = city.generic_energy_systems[energy_systems_cluster_name]
      for _generic_building_energy_system in _generic_building_energy_systems:
        _building_energy_systems.append(copy.deepcopy(_generic_building_energy_system))

      building.energy_systems = _building_energy_systems




class ListUsageToHub:
  """
  Eilat function to hub function class
  """

  def __init__(self, function_dictionary=None):
    self._function_dictionary = function_dictionary

  def _apply_function_dictionary(self, usages):

    function_dictionary = self._function_dictionary

    if function_dictionary is not None:
      for usage in usages:
        if usage['usage'] in function_dictionary:
          usage['usage'] = function_dictionary[usage['usage']]

    return usages
    
  def parse(self, usages) -> list[dict]:
    """
    Get the dictionary
    :return: {}
    """

    usages = [{"usage": str(i["usage"]), "ratio": float(i["ratio"])} for i in usages]

    usages = self._apply_function_dictionary(usages)

    return usages



"""
Dictionaries module for Palma function to hub function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Cecilia Pérez cperez@irec.cat
"""

import hub.helpers.constants as cte


class PalmaFunctionToHubFunction:
  """
  Palma function to hub function class
  """

  def __init__(self):
    self._dictionary = {'Residential': cte.RESIDENTIAL,
                        'Single-family building': cte.SINGLE_FAMILY_HOUSE,
                        'Large multifamily building': cte.HIGH_RISE_APARTMENT,
                        'Medium multifamily building': cte.MID_RISE_APARTMENT,
                        'Small multifamily building': cte.MULTI_FAMILY_HOUSE,
                        'V': cte.RESIDENTIAL
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary


"""
Insel monthly energy balance
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Project collaborator Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
import csv

from hub.helpers.monthly_values import MonthlyValues
import hub.helpers.constants as cte


class EnergyPlus:
  """
  Energy plus class
  """
  def __init__(self, city, base_path):
    self._city = city
    self._base_path = base_path

  @staticmethod
  def _building_energy_demands(energy_plus_output_file_path):
    with open(Path(energy_plus_output_file_path).resolve(), 'r', encoding='utf8') as csv_file:
      csv_output = csv.reader(csv_file)
      headers = next(csv_output)
      building_energy_demands = {
                                  'Heating (J)': [],
                                  'Cooling (J)': [],
                                  'DHW (J)': [],
                                  'Appliances (J)': [],
                                  'Lighting (J)': []
                                }
      heating_column_index = []
      cooling_column_index = []
      dhw_column_index = []
      appliance_column_index = []
      lighting_column_index = []
      for index, header in enumerate(headers):
        if "Total Heating" in header:
          heating_column_index.append(index)
        elif "Total Cooling" in header:
          cooling_column_index.append(index)
        elif "DHW" in header:
          dhw_column_index.append(index)
        elif "InteriorEquipment" in header:
          appliance_column_index.append(index)
        elif "InteriorLights" in header:
          lighting_column_index.append(index)

      for line in csv_output:
        total_heating_demand = 0
        total_cooling_demand = 0
        total_dhw_demand = 0
        total_appliance_demand = 0
        total_lighting_demand = 0
        for heating_index in heating_column_index:
          total_heating_demand += float(line[heating_index])
        building_energy_demands['Heating (J)'].append(total_heating_demand)
        for cooling_index in cooling_column_index:
          total_cooling_demand += float(line[cooling_index])
        building_energy_demands['Cooling (J)'].append(total_cooling_demand)
        for dhw_index in dhw_column_index:
          total_dhw_demand += float(line[dhw_index]) * 3600
        building_energy_demands['DHW (J)'].append(total_dhw_demand)
        for appliance_index in appliance_column_index:
          total_appliance_demand += float(line[appliance_index])
        building_energy_demands['Appliances (J)'].append(total_appliance_demand)
        for lighting_index in lighting_column_index:
          total_lighting_demand += float(line[lighting_index])
        building_energy_demands['Lighting (J)'].append(total_lighting_demand)

    return building_energy_demands

  def enrich(self):
    """
    Enrich the city by using the energy plus workflow output files (J)
    :return: None
    """
    for building in self._city.buildings:
      file_name = f'{building.name}_out.csv'
      energy_plus_output_file_path = Path(self._base_path / file_name).resolve()
      if energy_plus_output_file_path.is_file():
        building_energy_demands = self._building_energy_demands(energy_plus_output_file_path)
        building.heating_demand[cte.HOUR] = building_energy_demands['Heating (J)']
        building.cooling_demand[cte.HOUR] = building_energy_demands['Cooling (J)']
        building.domestic_hot_water_heat_demand[cte.HOUR] = building_energy_demands['DHW (J)']
        building.appliances_electrical_demand[cte.HOUR] = building_energy_demands['Appliances (J)']
        building.lighting_electrical_demand[cte.HOUR] = building_energy_demands['Lighting (J)']
        # todo: @Saeed, this a list of ONE value with the total energy of the year, exactly the same as cte.YEAR.
        #  You have to use the method to add hourly values from helpers/monthly_values
        building.heating_demand[cte.MONTH] = MonthlyValues.get_total_month(building.heating_demand[cte.HOUR])
        building.cooling_demand[cte.MONTH] = MonthlyValues.get_total_month(building.cooling_demand[cte.HOUR])
        building.domestic_hot_water_heat_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.domestic_hot_water_heat_demand[cte.HOUR]))
        building.appliances_electrical_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.appliances_electrical_demand[cte.HOUR]))
        building.lighting_electrical_demand[cte.MONTH] = (
          MonthlyValues.get_total_month(building.lighting_electrical_demand[cte.HOUR]))
        building.heating_demand[cte.YEAR] = [sum(building.heating_demand[cte.MONTH])]
        building.cooling_demand[cte.YEAR] = [sum(building.cooling_demand[cte.MONTH])]
        building.domestic_hot_water_heat_demand[cte.YEAR] = [sum(building.domestic_hot_water_heat_demand[cte.MONTH])]
        building.appliances_electrical_demand[cte.YEAR] = [sum(building.appliances_electrical_demand[cte.MONTH])]
        building.lighting_electrical_demand[cte.YEAR] = [sum(building.lighting_electrical_demand[cte.MONTH])]



"""
TestConstructionCatalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Contributors  Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from unittest import TestCase
from hub.catalog_factories.construction_catalog_factory import ConstructionCatalogFactory


class TestConstructionCatalog(TestCase):

  def test_nrel_catalog(self):
    catalog = ConstructionCatalogFactory('nrel').catalog
    catalog_categories = catalog.names()
    constructions = catalog.names('constructions')
    windows = catalog.names('windows')
    materials = catalog.names('materials')
    self.assertEqual(33, len(constructions['constructions']))
    self.assertEqual(5, len(windows['windows']))
    self.assertEqual(33, len(materials['materials']))
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')

  def test_nrcan_catalog(self):
    catalog = ConstructionCatalogFactory('nrcan').catalog
    catalog_categories = catalog.names()
    constructions = catalog.names('constructions')
    windows = catalog.names('windows')
    materials = catalog.names('materials')
    self.assertEqual(540, len(constructions['constructions']))
    self.assertEqual(96, len(windows['windows']))
    self.assertEqual(552, len(materials['materials']))
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')

  def test_eilat_catalog(self):
    catalog = ConstructionCatalogFactory('eilat').catalog
    catalog_categories = catalog.names()
    constructions = catalog.names('constructions')
    windows = catalog.names('windows')
    materials = catalog.names('materials')
    self.assertEqual(9, len(constructions['constructions']))
    self.assertEqual(3, len(windows['windows']))
    self.assertEqual(553, len(materials['materials']))
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')

  def test_palma_catalog(self):
    catalog = ConstructionCatalogFactory('palma').catalog
    catalog_categories = catalog.names()
    constructions = catalog.names('constructions')
    windows = catalog.names('windows')
    materials = catalog.names('materials')
    self.assertEqual(29, len(constructions['constructions']))
    self.assertEqual(9, len(windows['windows']))
    self.assertEqual(122, len(materials['materials']))
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')



"""
Dictionaries module for hub function to NREL construction function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import hub.helpers.constants as cte


class HubFunctionToNrelConstructionFunction:
  """
  Hub function to nrel construction function
  """

  def __init__(self):
    self._dictionary = {
      cte.RESIDENTIAL: 'residential',
      cte.SINGLE_FAMILY_HOUSE: 'residential',
      cte.MULTI_FAMILY_HOUSE: 'midrise apartment',
      cte.ROW_HOUSE: 'midrise apartment',
      cte.MID_RISE_APARTMENT: 'midrise apartment',
      cte.HIGH_RISE_APARTMENT: 'high-rise apartment',
      cte.OFFICE_AND_ADMINISTRATION: 'medium office',
      cte.SMALL_OFFICE: 'small office',
      cte.MEDIUM_OFFICE: 'medium office',
      cte.LARGE_OFFICE: 'large office',
      cte.COURTHOUSE: 'medium office',
      cte.FIRE_STATION: 'n/a',
      cte.PENITENTIARY: 'large hotel',
      cte.POLICE_STATION: 'n/a',
      cte.POST_OFFICE: 'medium office',
      cte.LIBRARY: 'medium office',
      cte.EDUCATION: 'secondary school',
      cte.PRIMARY_SCHOOL: 'primary school',
      cte.PRIMARY_SCHOOL_WITH_SHOWER: 'primary school',
      cte.SECONDARY_SCHOOL: 'secondary school',
      cte.UNIVERSITY: 'secondary school',
      cte.LABORATORY_AND_RESEARCH_CENTER: 'secondary school',
      cte.STAND_ALONE_RETAIL: 'stand-alone retail',
      cte.HOSPITAL: 'hospital',
      cte.OUT_PATIENT_HEALTH_CARE: 'outpatient healthcare',
      cte.HEALTH_CARE: 'outpatient healthcare',
      cte.RETIREMENT_HOME_OR_ORPHANAGE: 'small hotel',
      cte.COMMERCIAL: 'strip mall',
      cte.STRIP_MALL: 'strip mall',
      cte.SUPERMARKET: 'supermarket',
      cte.RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD: 'stand-alone retail',
      cte.RETAIL_SHOP_WITH_REFRIGERATED_FOOD: 'stand-alone retail',
      cte.RESTAURANT: 'full service restaurant',
      cte.QUICK_SERVICE_RESTAURANT: 'quick service restaurant',
      cte.FULL_SERVICE_RESTAURANT: 'full service restaurant',
      cte.HOTEL: 'small hotel',
      cte.HOTEL_MEDIUM_CLASS: 'small hotel',
      cte.SMALL_HOTEL: 'small hotel',
      cte.LARGE_HOTEL: 'large hotel',
      cte.DORMITORY: 'small hotel',
      cte.EVENT_LOCATION: 'n/a',
      cte.CONVENTION_CENTER: 'n/a',
      cte.HALL: 'n/a',
      cte.GREEN_HOUSE: 'n/a',
      cte.INDUSTRY: 'n/a',
      cte.WORKSHOP: 'n/a',
      cte.WAREHOUSE: 'warehouse',
      cte.WAREHOUSE_REFRIGERATED: 'warehouse',
      cte.SPORTS_LOCATION: 'n/a',
      cte.SPORTS_ARENA: 'n/a',
      cte.GYMNASIUM: 'n/a',
      cte.MOTION_PICTURE_THEATRE: 'n/a',
      cte.MUSEUM: 'n/a',
      cte.PERFORMING_ARTS_THEATRE: 'n/a',
      cte.TRANSPORTATION: 'n/a',
      cte.AUTOMOTIVE_FACILITY: 'n/a',
      cte.PARKING_GARAGE: 'n/a',
      cte.RELIGIOUS: 'n/a',
      cte.NON_HEATED: 'n/a',
      cte.DATACENTER: 'n/a',
      cte.FARM: 'n/a'
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
EpwWeatherParameters class to extract weather parameters from a defined region in .epw format (EnergyPlus Weather)
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
import sys
from pathlib import Path
import requests
import pandas as pd

import hub.helpers.constants as cte
from hub.helpers.monthly_values import MonthlyValues
from hub.city_model_structure.city import City
from hub.imports.weather.helpers.weather import Weather as wh


class EpwWeatherParameters:
  """
  EpwWeatherParameters class
  """
  def __init__(self, city: City):
    self._weather_values = None
    self._city = city
    self._url = wh().epw_file(city.region_code)
    self._path = (Path(__file__).parent.parent.parent / f'data/weather/epw/{self._url.rsplit("/", 1)[1]}').resolve()
    if not self._path.exists():
      with open(self._path, 'wb') as epw_file:
        epw_file.write(requests.get(self._url, allow_redirects=True).content)

    try:
      with open(self._path, 'r', encoding='utf8') as file:
        line = file.readline().split(',')
        city.climate_reference_city = line[1]
        city.latitude = line[6]
        city.longitude = line[7]
        city.time_zone = line[8]
        for i in range(0, 2):
          _ = file.readline().split(',')
        line = file.readline().split(',')
        number_records = int(line[1])
        ground_temperature_from_file = {}
        for i in range(0, number_records):
          depth_measurement_ground_temperature = line[i*16+2]
          temperatures = []
          for j in range(0, 12):
            temperatures.append(float(line[i*16+j+6]))
          ground_temperature_from_file[depth_measurement_ground_temperature] = temperatures
    except SystemExit:
      logging.error('Error: weather file %s not found. Please download it from https://energyplus.net/weather and place'
                    ' it in folder data\\weather\\epw', self._path)
      sys.exit()

    try:
      self._weather_values = pd.read_csv(self._path, header=0, skiprows=7,
                                         names=['year', 'month', 'day', 'hour', 'minute',
                                                'data_source_and_uncertainty_flags',
                                                'dry_bulb_temperature_c', 'dew_point_temperature_c',
                                                'relative_humidity_perc',
                                                'atmospheric_station_pressure_pa',
                                                'extraterrestrial_horizontal_radiation_wh_m2',
                                                'extraterrestrial_direct_normal_radiation_wh_m2',
                                                'horizontal_infrared_radiation_intensity_wh_m2',
                                                'global_horizontal_radiation_wh_m2', 'direct_normal_radiation_wh_m2',
                                                'diffuse_horizontal_radiation_wh_m2',
                                                'global_horizontal_illuminance_lux',
                                                'direct_normal_illuminance_lux', 'diffuse_horizontal_illuminance_lux',
                                                'zenith_luminance_cd_m2',
                                                'wind_direction_deg', 'wind_speed_m_s', 'total_sky_cover',
                                                'opaque_sky_cover', 'visibility_km',
                                                'ceiling_heigh_m_s', 'present_weather_observation',
                                                'present_weather_codes',
                                                'precipitable_water_mm', 'aerosol_optical_depth_10_3_ths',
                                                'snow_depth_cm',
                                                'days_since_last_snowfall', 'albedo', 'liquid_precipitation_depth_mm',
                                                'liquid_precipitation_quality_hr'])
      number_invalid_records = self._weather_values[
        self._weather_values.dry_bulb_temperature_c == 99.9].count().dry_bulb_temperature_c
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 99.9) in dry bulb temperature\n')
      number_invalid_records = self._weather_values[
        self._weather_values.global_horizontal_radiation_wh_m2 == 9999].count().global_horizontal_radiation_wh_m2
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in global horizontal radiation\n')
      number_invalid_records = self._weather_values[
        self._weather_values.diffuse_horizontal_radiation_wh_m2 == 9999].count().diffuse_horizontal_radiation_wh_m2
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in diffuse horizontal radiation\n')
      number_invalid_records = self._weather_values[
        self._weather_values.direct_normal_radiation_wh_m2 == 9999].count().direct_normal_radiation_wh_m2
      if number_invalid_records > 0:
        sys.stderr.write(f'Warning: {self._path} invalid records (value of 9999) in direct horizontal radiation\n')

      self._weather_values = self._weather_values.to_dict(orient='list')

    except SystemExit:
      sys.stderr.write(f'Error: wrong formatting of weather file {self._path}\n')
      sys.exit()
    for building in self._city.buildings:
      building.ground_temperature[cte.MONTH] = ground_temperature_from_file
      ground_temperature = {}
      for ground_temperature_set in building.ground_temperature[cte.MONTH]:
        temperature = sum(building.ground_temperature[cte.MONTH][ground_temperature_set]) / 12
        ground_temperature[ground_temperature_set] = [temperature]
      building.ground_temperature[cte.YEAR] = ground_temperature
      if cte.HOUR in building.external_temperature:
        del building.external_temperature[cte.HOUR]
#      new_value = pd.DataFrame(self._weather_values[['dry_bulb_temperature_c']].to_numpy(), columns=['epw'])
#      number_invalid_records = new_value[new_value.epw == 99.9].count().epw
      building.external_temperature[cte.HOUR] = self._weather_values['dry_bulb_temperature_c']
      building.global_horizontal[cte.HOUR] = [x for x in self._weather_values['global_horizontal_radiation_wh_m2']]
      building.diffuse[cte.HOUR] = [x for x in self._weather_values['diffuse_horizontal_radiation_wh_m2']]
      building.direct_normal[cte.HOUR] = [x for x in self._weather_values['direct_normal_radiation_wh_m2']]
      building.beam[cte.HOUR] = [building.global_horizontal[cte.HOUR][i] -
                                 building.diffuse[cte.HOUR][i]
                                 for i in range(len(building.global_horizontal[cte.HOUR]))]
      building.cold_water_temperature[cte.HOUR] = wh().cold_water_temperature(building.external_temperature[cte.HOUR])


    # create the monthly and yearly values out of the hourly
    for building in self._city.buildings:
      building.external_temperature[cte.MONTH] = \
        MonthlyValues().get_mean_values(building.external_temperature[cte.HOUR])
      building.external_temperature[cte.YEAR] = [sum(building.external_temperature[cte.HOUR]) / 8760]
      building.cold_water_temperature[cte.MONTH] = \
        MonthlyValues().get_mean_values(building.cold_water_temperature[cte.HOUR])
      building.cold_water_temperature[cte.YEAR] = [sum(building.cold_water_temperature[cte.HOUR]) / 8760]

      # If the usage has already being imported, the domestic hot water missing values must be calculated here that
      # the cold water temperature is finally known
      cold_temperature = building.cold_water_temperature[cte.YEAR][0]
      for internal_zone in building.internal_zones:
        if internal_zone.usages is not None:
          for usage in internal_zone.usages:
            if usage.domestic_hot_water.peak_flow is None:
              if usage.domestic_hot_water.density is None:
                continue
              peak_flow = 0
              if (usage.domestic_hot_water.service_temperature - cold_temperature) > 0:
                peak_flow = usage.domestic_hot_water.density / cte.WATER_DENSITY / cte.WATER_HEAT_CAPACITY \
                      / (usage.domestic_hot_water.service_temperature - cold_temperature)
              usage.domestic_hot_water.peak_flow = peak_flow
            if usage.domestic_hot_water.density is None:
              if usage.domestic_hot_water.peak_flow is None:
                continue
              density = (
                  usage.domestic_hot_water.peak_flow * cte.WATER_DENSITY * cte.WATER_HEAT_CAPACITY *
                  (usage.domestic_hot_water.service_temperature - cold_temperature)
              )
              usage.domestic_hot_water.density = density

    self._city.level_of_detail.weather = 2
    for building in self._city.buildings:
      building.level_of_detail.weather = 2



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from abc import ABC


class EnergyStorageSystem(ABC):
  """"
  Energy Storage System Abstract Class
  """

  def __init__(self, storage_id, model_name=None, manufacturer=None,
               nominal_capacity=None, losses_ratio=None):
    self._storage_id = storage_id
    self._model_name = model_name
    self._manufacturer = manufacturer
    self._nominal_capacity = nominal_capacity
    self._losses_ratio = losses_ratio

  @property
  def id(self):
    """
    Get storage id
    :return: string
    """
    return self._storage_id

  @property
  def type_energy_stored(self):
    """
    Get type of energy stored from ['electrical', 'thermal']
    :return: string
    """
    raise NotImplementedError

  @property
  def model_name(self):
    """
    Get system model
    :return: string
    """
    return self._model_name

  @property
  def manufacturer(self):
    """
    Get name of manufacturer
    :return: string
    """
    return self._manufacturer

  @property
  def nominal_capacity(self):
    """
    Get the nominal capacity of the storage system in Jules
    :return: float
    """
    return self._nominal_capacity

  @property
  def losses_ratio(self):
    """
    Get the losses-ratio of storage system in Jules lost / Jules stored
    :return: float
    """
    return self._losses_ratio

  def to_dictionary(self):
    """Class content to dictionary"""
    raise NotImplementedError



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations


class PerformanceCurves:
  """
  Parameter function class
  """

  def __init__(self):
    self._curve_type = None
    self._dependant_variable = None
    self._parameters = None
    self._coefficients = None

  @property
  def curve_type(self):
    """
    Get the type of the fit function from the following
    Linear =>>> y = a + b*x
    Exponential =>>> y = a*(b**x)
    Second degree polynomial =>>> y = a + b*x + c*(x**2)
    Power =>>> y = a*(x**b)
    Bi-Quadratic =>>> y = a + b*x + c*(x**2) + d*z + e*(z**2) + f*x*z

    Get the type of function from ['linear', 'exponential', 'second degree polynomial', 'power', 'bi-quadratic']
    :return: string
    """

    return self._curve_type

  @curve_type.setter
  def curve_type(self, value):
    """
    Set the type of the fit function from the following
    Linear =>>> y = a + b*x
    Exponential =>>> y = a*(b**x)
    Second degree polynomial =>>> y = a + b*x + c*(x**2)
    Power =>>> y = a*(x**b)
    Bi-Quadratic =>>> y = a + b*x + c*(x**2) + d*z + e*(z**2) + f*x*z

    Get the type of function from ['linear', 'exponential', 'second degree polynomial', 'power', 'bi-quadratic']
    :return: string
    """
    self._curve_type = value

  @property
  def dependant_variable(self):
    """
    Get y (e.g. COP in COP = a*source temperature**2 + b*source temperature + c*source temperature*supply temperature +
    d*supply temperature + e*supply temperature**2 + f)
    """
    return self._dependant_variable

  @dependant_variable.setter
  def dependant_variable(self, value):
    """
    Set y (e.g. COP in COP = a*source temperature**2 + b*source temperature + c*source temperature*supply temperature +
    d*supply temperature + e*supply temperature**2 + f)
    """
    self._dependant_variable = value

  @property
  def parameters(self):
    """
    Get the list of parameters involved in fitting process as ['x', 'z'] (e.g. [source temperature, supply temperature]
    in COP= *source temperature**2 + b*source temperature + c*source temperature*supply temperature +
    d*supply temperature + e*supply temperature**2 + f)
    :return: string
    """
    return self._parameters

  @parameters.setter
  def parameters(self, value):
    """
    Set the list of parameters involved in fitting process as ['x', 'z'] (e.g. [source temperature, supply temperature]
    in COP= *source temperature**2 + b*source temperature + c*source temperature*supply temperature +
    d*supply temperature + e*supply temperature**2 + f)
    :return: string
    """
    self._parameters = value

  @property
  def coefficients(self):
    """
    Get the coefficients of the functions as list of ['a', 'b', 'c', 'd', 'e', 'f']
    :return: [coefficients]
    """
    return self._coefficients

  @coefficients.setter
  def coefficients(self, value):
    """
    Set the coefficients of the functions as list of ['a', 'b', 'c', 'd', 'e', 'f']
    :return: [coefficients]
    """
    self._coefficients = value



"""
EilatPhysicsParameters import the construction and material information defined for Eilat
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging

from hub.catalog_factories.construction_catalog_factory import ConstructionCatalogFactory
from hub.city_model_structure.building_demand.thermal_archetype import ThermalArchetype
from hub.city_model_structure.building_demand.construction import Construction
from hub.city_model_structure.building_demand.layer import Layer
from hub.helpers.dictionaries import Dictionaries
from hub.imports.construction.helpers.construction_helper import ConstructionHelper


class EilatPhysicsParameters:
  """
  EilatPhysicsParameters class
  """

  def __init__(self, city, divide_in_storeys=False):
    self._city = city
    self._divide_in_storeys = divide_in_storeys
    self._climate_zone = ConstructionHelper.city_to_israel_climate_zone(city.climate_reference_city)

  def enrich_buildings(self):
    """
    Returns the city with the construction parameters assigned to the buildings
    """
    city = self._city
    eilat_catalog = ConstructionCatalogFactory('eilat').catalog
    for building in city.buildings:
      if building.function not in Dictionaries().hub_function_to_eilat_construction_function:
        logging.error('Building %s has an unknown building function %s', building.name, building.function)
        continue
      function = Dictionaries().hub_function_to_eilat_construction_function[building.function]
      try:
        archetype = self._search_archetype(eilat_catalog, function, building.year_of_construction, self._climate_zone)

      except KeyError:
        logging.error('Building %s has unknown construction archetype for building function: %s '
                      '[%s], building year of construction: %s and climate zone %s', building.name, function,
                      building.function, building.year_of_construction, self._climate_zone)
        continue
      thermal_archetype = ThermalArchetype()
      self._assign_values(thermal_archetype, archetype)
      for internal_zone in building.internal_zones:
        internal_zone.thermal_archetype = thermal_archetype

  @staticmethod
  def _search_archetype(nrcan_catalog, function, year_of_construction, climate_zone):
    nrcan_archetypes = nrcan_catalog.entries('archetypes')
    for building_archetype in nrcan_archetypes:
      construction_period_limits = building_archetype.construction_period.split('_')
      if int(construction_period_limits[0]) <= int(year_of_construction) <= int(construction_period_limits[1]):
        if str(function) == str(building_archetype.function) and climate_zone == str(building_archetype.climate_zone):
          return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _assign_values(thermal_archetype, catalog_archetype):
    thermal_archetype.average_storey_height = catalog_archetype.average_storey_height
    thermal_archetype.extra_loses_due_to_thermal_bridges = catalog_archetype.extra_loses_due_to_thermal_bridges
    thermal_archetype.indirect_heated_ratio = 0
    thermal_archetype.infiltration_rate_for_ventilation_system_on = catalog_archetype.infiltration_rate_for_ventilation_system_on
    thermal_archetype.infiltration_rate_for_ventilation_system_off = catalog_archetype.infiltration_rate_for_ventilation_system_off
    effective_thermal_capacity = 0
    _constructions = []
    for catalog_construction in catalog_archetype.constructions:
      construction = Construction()
      construction.type = catalog_construction.type
      construction.name = catalog_construction.name
      if catalog_construction.window_ratio is not None:
        for _orientation in catalog_construction.window_ratio:
          if catalog_construction.window_ratio[_orientation] is None:
            catalog_construction.window_ratio[_orientation] = 0
      construction.window_ratio = catalog_construction.window_ratio
      _layers = []
      total_thickness = 0
      for layer_archetype in catalog_construction.layers:
        layer = Layer()
        layer.thickness = layer_archetype.thickness
        total_thickness += layer_archetype.thickness
        archetype_material = layer_archetype.material
        layer.material_name = archetype_material.name
        layer.no_mass = archetype_material.no_mass
        if archetype_material.no_mass:
          layer.thermal_resistance = archetype_material.thermal_resistance
        else:
          layer.density = archetype_material.density
          layer.conductivity = archetype_material.conductivity
          layer.specific_heat = archetype_material.specific_heat
          effective_thermal_capacity += archetype_material.specific_heat \
                                        * archetype_material.density * layer_archetype.thickness
        layer.solar_absorptance = archetype_material.solar_absorptance
        layer.thermal_absorptance = archetype_material.thermal_absorptance
        layer.visible_absorptance = archetype_material.visible_absorptance
        _layers.append(layer)
      construction.layers = _layers
      effective_thermal_capacity = effective_thermal_capacity / total_thickness

      if catalog_construction.window is not None:
        window_archetype = catalog_construction.window
        construction.window_frame_ratio = window_archetype.frame_ratio
        construction.window_g_value = window_archetype.g_value
        construction.window_overall_u_value = window_archetype.overall_u_value
      _constructions.append(construction)
    thermal_archetype.constructions = _constructions
    thermal_archetype.thermal_capacity = effective_thermal_capacity



"""
export a city into Glb format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import os
import shutil
import subprocess

from hub.city_model_structure.city import City
from hub.exports.formats.obj import Obj


class GltExceptionError(Exception):
  """
  Glt execution error
  """


class Glb:
  """
  Glb class
  """
  def __init__(self, city, path, target_buildings=None):
    self._city = city
    self._path = path
    if target_buildings is None:
      target_buildings = [b.name for b in self._city.buildings]
    self._target_buildings = target_buildings
    self._export()

  @property
  def _obj2gltf(self):
    return shutil.which('obj2gltf')

  def _export(self):
    try:
      for building in self._city.buildings:
        city = City(self._city.lower_corner, self._city.upper_corner, self._city.srs_name)
        city.add_city_object(building)
        city.name = building.name
        Obj(city, self._path)
        glb = f'{self._path}/{building.name}.glb'
        subprocess.run([
          self._obj2gltf,
          '-i', f'{self._path}/{building.name}.obj',
          '-b',
          '-o', f'{glb}'
        ])
        os.unlink(f'{self._path}/{building.name}.obj')
        os.unlink(f'{self._path}/{building.name}.mtl')
    except (subprocess.SubprocessError, subprocess.TimeoutExpired, subprocess.CalledProcessError) as err:
      raise GltExceptionError from err



"""
Energy System catalog heat generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Saeed Ranjbar saeed.ranjbar@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations


class PerformanceCurves:
  """
  Parameter function class
  """

  def __init__(self, curve_type, dependant_variable, parameters, coefficients):
    self._curve_type = curve_type
    self._dependant_variable = dependant_variable
    self._parameters = parameters
    self._coefficients = coefficients

  @property
  def curve_type(self):
    """
    The type of the fit function from the following
    Linear =>>> y = a + b*x
    Exponential =>>> y = a*(b**x)
    Second degree polynomial =>>> y = a + b*x + c*(x**2)
    Power =>>> y = a*(x**b)
    Bi-Quadratic =>>> y = a + b*x + c*(x**2) + d*z + e*(z**2) + f*x*z

    Get the type of function from ['linear', 'exponential', 'second degree polynomial', 'power', 'bi-quadratic']
    :return: string
    """
    return self._curve_type

  @property
  def dependant_variable(self):
    """
    y (e.g. COP in COP = a*source temperature**2 + b*source temperature + c*source temperature*supply temperature +
    d*supply temperature + e*supply temperature**2 + f)
    """
    return self._dependant_variable

  @property
  def parameters(self):
    """
    Get the list of parameters involved in fitting process as ['x', 'z'] (e.g. [source temperature, supply temperature]
    in COP=)
    :return: string
    """
    return self._parameters

  @property
  def coefficients(self):
    """
    Get the coefficients of the functions as list of ['a', 'b', 'c', 'd', 'e', 'f']
    :return: [coefficients]
    """
    return self._coefficients

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Parameter Function': {
      'curve type': self.curve_type,
      'dependant variable': self.dependant_variable,
      'parameter(s)': self.parameters,
      'coefficients': self.coefficients,
    }
    }
    return content



"""
ComnetUsageParameters extracts the usage properties from Comnet catalog and assigns to each building
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Project Collaborator Saeed Ranjbar saeed.ranjbar@concordia.ca
"""
import copy
import logging
import numpy
import hub.helpers.constants as cte
from hub.helpers.dictionaries import Dictionaries
from hub.city_model_structure.building_demand.usage import Usage
from hub.city_model_structure.building_demand.lighting import Lighting
from hub.city_model_structure.building_demand.occupancy import Occupancy
from hub.city_model_structure.building_demand.appliances import Appliances
from hub.city_model_structure.building_demand.thermal_control import ThermalControl
from hub.city_model_structure.building_demand.domestic_hot_water import DomesticHotWater
from hub.city_model_structure.attributes.schedule import Schedule
from hub.city_model_structure.building_demand.internal_gain import InternalGain
from hub.catalog_factories.usage_catalog_factory import UsageCatalogFactory
from hub.catalog_factories.construction_catalog_factory import ConstructionCatalogFactory
from hub.imports.construction.helpers.construction_helper import ConstructionHelper


class ComnetUsageParameters:
  """
  ComnetUsageParameters class
  """
  def __init__(self, city):
    self._city = city

  def enrich_buildings(self):
    """
    Returns the city with the usage parameters assigned to the buildings
    :return:
    """
    city = self._city
    comnet_catalog = UsageCatalogFactory('comnet').catalog
    for building in city.buildings:
      comnet_archetype_usages = []
      usages = building.usages
      for usage in usages:
        comnet_usage_name = Dictionaries().hub_usage_to_comnet_usage[usage['usage']]
        try:
          comnet_archetype_usage = self._search_archetypes(comnet_catalog, comnet_usage_name)
          comnet_archetype_usages.append(comnet_archetype_usage)
        except KeyError:
          logging.error('Building %s has unknown usage archetype for usage %s', building.name, comnet_usage_name)
          continue

      for (i, internal_zone) in enumerate(building.internal_zones):
        internal_zone_usages = []
        if len(building.internal_zones) > 1:
          volume_per_area = 0
          if internal_zone.area is None:
            logging.error('Building %s has internal zone area not defined, ACH cannot be calculated for usage %s',
                          building.name, usages[i]['usage'])
            continue
          if internal_zone.volume is None:
            logging.error('Building %s has internal zone volume not defined, ACH cannot be calculated for usage %s',
                          building.name, usages[i]['usage'])
            continue
          if internal_zone.area <= 0:
            logging.error('Building %s has internal zone area equal to 0, ACH cannot be calculated for usage %s',
                          building.name,  usages[i]['usage'])
            continue
          volume_per_area += internal_zone.volume / internal_zone.area
          usage = Usage()
          usage.name = usages[i]['usage']
          self._assign_values(usage, comnet_archetype_usages[i], volume_per_area, building.cold_water_temperature)
          usage.percentage = 1
          self._calculate_reduced_values_from_extended_library(usage, comnet_archetype_usages[i])
          internal_zone_usages.append(usage)
        else:
          storeys_above_ground = building.storeys_above_ground
          if storeys_above_ground is None:
            logging.error('Building %s no number of storeys assigned, ACH cannot be calculated for usage %s. '
                          'NRCAN construction data for the year %s is used to calculated number of storeys above '
                          'ground', building.name,  usages, building.year_of_construction)
            try:
              storeys_above_ground = self.average_storey_height_calculator(self._city, building)
            except ValueError as e:
              logging.error(e)
              continue

          volume_per_area = building.volume / building.floor_area / storeys_above_ground
          for j, usage_type in enumerate(usages):
            usage = Usage()
            usage.name = usage_type['usage']
            usage.percentage = float(usage_type['ratio'])

            self._assign_values(usage, comnet_archetype_usages[j], volume_per_area, building.cold_water_temperature)
            self._calculate_reduced_values_from_extended_library(usage, comnet_archetype_usages[j])
            internal_zone_usages.append(usage)

        internal_zone.usages = internal_zone_usages

  @staticmethod
  def _search_archetypes(comnet_catalog, usage_name):
    comnet_archetypes = comnet_catalog.entries('archetypes').usages
    for building_archetype in comnet_archetypes:
      if str(usage_name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _assign_values(usage, archetype, volume_per_area, cold_water_temperature):
    # Due to the fact that python is not a typed language, the wrong object type is assigned to
    # usage.occupancy when writing usage.occupancy = archetype.occupancy.
    # Same happens for lighting and appliances. Therefore, this walk around has been done.
    usage.mechanical_air_change = archetype.ventilation_rate / volume_per_area
    _occupancy = Occupancy()
    _occupancy.occupancy_density = archetype.occupancy.occupancy_density
    _occupancy.sensible_radiative_internal_gain = archetype.occupancy.sensible_radiative_internal_gain
    _occupancy.latent_internal_gain = archetype.occupancy.latent_internal_gain
    _occupancy.sensible_convective_internal_gain = archetype.occupancy.sensible_convective_internal_gain
    _occupancy.occupancy_schedules = archetype.occupancy.schedules
    usage.occupancy = _occupancy
    _lighting = Lighting()
    _lighting.density = archetype.lighting.density
    _lighting.convective_fraction = archetype.lighting.convective_fraction
    _lighting.radiative_fraction = archetype.lighting.radiative_fraction
    _lighting.latent_fraction = archetype.lighting.latent_fraction
    _lighting.schedules = archetype.lighting.schedules
    usage.lighting = _lighting
    _appliances = Appliances()
    _appliances.density = archetype.appliances.density
    _appliances.convective_fraction = archetype.appliances.convective_fraction
    _appliances.radiative_fraction = archetype.appliances.radiative_fraction
    _appliances.latent_fraction = archetype.appliances.latent_fraction
    _appliances.schedules = archetype.appliances.schedules
    usage.appliances = _appliances
    _control = ThermalControl()
    _control.cooling_set_point_schedules = archetype.thermal_control.cooling_set_point_schedules
    _control.heating_set_point_schedules = archetype.thermal_control.heating_set_point_schedules
    _control.hvac_availability_schedules = archetype.thermal_control.hvac_availability_schedules
    usage.thermal_control = _control
    _domestic_hot_water = DomesticHotWater()
    _domestic_hot_water.density = archetype.domestic_hot_water.density
    _domestic_hot_water.service_temperature = archetype.domestic_hot_water.service_temperature
    peak_flow = None
    if len(cold_water_temperature) > 0:
      cold_temperature = cold_water_temperature[cte.YEAR][0]
      peak_flow = 0
      if (archetype.domestic_hot_water.service_temperature - cold_temperature) > 0:
        peak_flow = archetype.domestic_hot_water.density / cte.WATER_DENSITY / cte.WATER_HEAT_CAPACITY \
                    / (archetype.domestic_hot_water.service_temperature - cold_temperature)
    _domestic_hot_water.peak_flow = peak_flow
    _domestic_hot_water.schedules = archetype.domestic_hot_water.schedules
    usage.domestic_hot_water = _domestic_hot_water

  @staticmethod
  def _calculate_reduced_values_from_extended_library(usage, archetype):
    number_of_days_per_type = {'WD': 251, 'Sat': 52, 'Sun': 62}
    total = 0
    for schedule in archetype.thermal_control.hvac_availability_schedules:
      if schedule.day_types[0] == cte.SATURDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sat']
      elif schedule.day_types[0] == cte.SUNDAY:
        for value in schedule.values:
          total += value * number_of_days_per_type['Sun']
      else:
        for value in schedule.values:
          total += value * number_of_days_per_type['WD']

    usage.hours_day = total / 365
    usage.days_year = 365

    max_heating_setpoint = cte.MIN_FLOAT
    min_heating_setpoint = cte.MAX_FLOAT

    for schedule in archetype.thermal_control.heating_set_point_schedules:
      if schedule.values is None:
        max_heating_setpoint = None
        min_heating_setpoint = None
        break
      if max(schedule.values) > max_heating_setpoint:
        max_heating_setpoint = max(schedule.values)
      if min(schedule.values) < min_heating_setpoint:
        min_heating_setpoint = min(schedule.values)

    min_cooling_setpoint = cte.MAX_FLOAT
    for schedule in archetype.thermal_control.cooling_set_point_schedules:
      if schedule.values is None:
        min_cooling_setpoint = None
        break
      if min(schedule.values) < min_cooling_setpoint:
        min_cooling_setpoint = min(schedule.values)

    usage.thermal_control.mean_heating_set_point = max_heating_setpoint
    usage.thermal_control.heating_set_back = min_heating_setpoint
    usage.thermal_control.mean_cooling_set_point = min_cooling_setpoint

  @staticmethod
  def _calculate_internal_gains(archetype):

    _days = [cte.MONDAY, cte.TUESDAY, cte.WEDNESDAY, cte.THURSDAY, cte.FRIDAY, cte.SATURDAY, cte.SUNDAY, cte.HOLIDAY]
    _number_of_days_per_type = [51, 50, 50, 50, 50, 52, 52, 10]

    _mean_internal_gain = InternalGain()
    _mean_internal_gain.type = 'mean_value_of_internal_gains'
    _base_schedule = Schedule()
    _base_schedule.type = cte.INTERNAL_GAINS
    _base_schedule.time_range = cte.DAY
    _base_schedule.time_step = cte.HOUR
    _base_schedule.data_type = cte.FRACTION

    _latent_heat_gain = archetype.occupancy.latent_internal_gain
    _convective_heat_gain = archetype.occupancy.sensible_convective_internal_gain
    _radiative_heat_gain = archetype.occupancy.sensible_radiative_internal_gain
    _total_heat_gain = _latent_heat_gain + _convective_heat_gain + _radiative_heat_gain

    _schedule_values = numpy.zeros([24, 8])
    _sum = 0
    for day, _schedule in enumerate(archetype.occupancy.schedules):
      for v_index, value in enumerate(_schedule.values):
        _schedule_values[v_index, day] = value * _total_heat_gain
        _sum += value * _total_heat_gain * _number_of_days_per_type[day]

    _total_heat_gain += archetype.lighting.density + archetype.appliances.density
    _latent_heat_gain += (
        archetype.lighting.latent_fraction * archetype.lighting.density + archetype.appliances.latent_fraction *
        archetype.appliances.density
    )
    _radiative_heat_gain = (
        archetype.lighting.radiative_fraction * archetype.lighting.density + archetype.appliances.radiative_fraction *
        archetype.appliances.density
    )
    _convective_heat_gain = (
      archetype.lighting.convective_fraction * archetype.lighting.density + archetype.appliances.convective_fraction *
      archetype.appliances.density
    )

    for day, _schedule in enumerate(archetype.lighting.schedules):
      for v_index, value in enumerate(_schedule.values):
        _schedule_values[v_index, day] += value * archetype.lighting.density
        _sum += value * archetype.lighting.density * _number_of_days_per_type[day]

    for day, _schedule in enumerate(archetype.appliances.schedules):
      for v_index, value in enumerate(_schedule.values):
        _schedule_values[v_index, day] += value * archetype.appliances.density
        _sum += value * archetype.appliances.density * _number_of_days_per_type[day]

    _latent_fraction = 0
    _radiative_fraction = 0
    _convective_fraction = 0
    _average_internal_gain = 0
    if _total_heat_gain != 0:
      _latent_fraction = _latent_heat_gain / _total_heat_gain
      _radiative_fraction = _radiative_heat_gain / _total_heat_gain
      _convective_fraction = _convective_heat_gain / _total_heat_gain
      _average_internal_gain = _sum / _total_heat_gain

    _schedules = []
    for day, current_day in enumerate(_days):
      _schedule = copy.deepcopy(_base_schedule)
      _schedule.day_types = [current_day]
      _schedule.values = _schedule_values[:day]
      _schedules.append(_schedule)

    _mean_internal_gain.average_internal_gain = _average_internal_gain
    _mean_internal_gain.latent_fraction = _latent_fraction
    _mean_internal_gain.convective_fraction = _convective_fraction
    _mean_internal_gain.radiative_fraction = _radiative_fraction
    _mean_internal_gain.schedules = _schedules

    return [_mean_internal_gain]

  @staticmethod
  def average_storey_height_calculator(city, building):
    climate_zone = ConstructionHelper.city_to_nrcan_climate_zone(city.climate_reference_city)
    nrcan_catalog = ConstructionCatalogFactory('nrcan').catalog

    if building.function not in Dictionaries().hub_function_to_nrcan_construction_function:
      raise ValueError('Building %s has an unknown building function %s', building.name, building.function)

    function = Dictionaries().hub_function_to_nrcan_construction_function[building.function]
    construction_archetype = None
    average_storey_height = None
    nrcan_archetypes = nrcan_catalog.entries('archetypes')
    for building_archetype in nrcan_archetypes:
      construction_period_limits = building_archetype.construction_period.split('_')
      if int(construction_period_limits[0]) <= int(building.year_of_construction) <= int(construction_period_limits[1]):
        if str(function) == str(building_archetype.function) and climate_zone == str(building_archetype.climate_zone):
          construction_archetype = building_archetype
          average_storey_height = building_archetype.average_storey_height
    if construction_archetype is None:
      raise ValueError('Building %s has unknown construction archetype for building function: %s '
                    '[%s], building year of construction: %s and climate zone %s', building.name, function,
                    building.function, building.year_of_construction, climate_zone)

    return average_storey_height



"""
UsageFactory retrieve the specific usage module for the given region
This factory can only be called after calling the construction factory so the thermal zones are created.
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from hub.helpers.utils import validate_import_export_type
from hub.imports.usage.comnet_usage_parameters import ComnetUsageParameters
from hub.imports.usage.nrcan_usage_parameters import NrcanUsageParameters
from hub.imports.usage.eilat_usage_parameters import EilatUsageParameters
from hub.imports.usage.palma_usage_parameters import PalmaUsageParameters


class UsageFactory:
  """
  UsageFactory class
  """
  def __init__(self, handler, city):
    self._handler = '_' + handler.lower()
    validate_import_export_type(UsageFactory, handler)
    self._city = city

  def _comnet(self):
    """
    Enrich the city with COMNET usage library
    """
    ComnetUsageParameters(self._city).enrich_buildings()
    self._city.level_of_detail.usage = 2
    for building in self._city.buildings:
      building.level_of_detail.usage = 2

  def _nrcan(self):
    """
    Enrich the city with NRCAN usage library
    """
    NrcanUsageParameters(self._city).enrich_buildings()
    self._city.level_of_detail.usage = 2
    for building in self._city.buildings:
      building.level_of_detail.usage = 2

  def _eilat(self):
    """
    Enrich the city with Eilat usage library
    """
    EilatUsageParameters(self._city).enrich_buildings()
    self._city.level_of_detail.usage = 2
    for building in self._city.buildings:
      building.level_of_detail.usage = 2

  def _palma(self):
    """
    Enrich the city with Palma usage library
    """
    PalmaUsageParameters(self._city).enrich_buildings()
    self._city.level_of_detail.usage = 2
    for building in self._city.buildings:
      building.level_of_detail.usage = 2

  def enrich(self):
    """
    Enrich the city given to the class using the usage factory given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()





"""
export a city into Obj format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from pathlib import Path

import numpy as np


class Obj:
  """
  Export to obj format
  """
  def __init__(self, city, path):
    self._city = city
    self._path = path
    self._export()

  def _ground(self, coordinate):
    x = coordinate[0] - self._city.lower_corner[0]
    y = coordinate[1] - self._city.lower_corner[1]
    z = coordinate[2] - self._city.lower_corner[2]
    return x, y, z

  def _to_vertex(self, coordinate):
    x, y, z = self._ground(coordinate)
    return f'v {x} {z} -{y}\n'  # to match opengl expectations

  def _to_texture_vertex(self, coordinate):
    u, v, _ = self._ground(coordinate)
    return f'vt {u} {v}\n'

  def _to_normal_vertex(self, coordinates):
    ground_vertex = []
    for coordinate in coordinates:
      x, y, z = self._ground(coordinate)
      ground_vertex.append(np.array([x, y, z]))
    # recalculate the normal to get grounded values
    edge_1 = ground_vertex[1] - ground_vertex[0]
    edge_2 = ground_vertex[2] - ground_vertex[0]
    normal = np.cross(edge_1, edge_2)
    normal = normal / np.linalg.norm(normal)
    return f'vn {normal[0]} {normal[1]} {normal[2]}\n'

  def _export(self):
    if self._city.name is None:
      self._city.name = 'unknown_city'
    obj_name = f'{self._city.name}.obj'
    mtl_name = f'{self._city.name}.mtl'
    obj_file_path = (Path(self._path).resolve() / obj_name).resolve()
    mtl_file_path = (Path(self._path).resolve() / mtl_name).resolve()
    with open(mtl_file_path, 'w', encoding='utf-8') as mtl:
      mtl.write("newmtl cerc_base_material\n")
      mtl.write("Ka 1.0 1.0 1.0      # Ambient color (white)\n")
      mtl.write("Kd 0.1 0.3 0.1      # Diffuse color (greenish)\n")
      mtl.write("Ks 1.0 1.0 1.0      # Specular color (white)\n")
      mtl.write("Ns 400.0             # Specular exponent (defines shininess)\n")
    vertices = {}
    faces = []
    vertex_index = 0
    normal_index = 0
    with open(obj_file_path, 'w', encoding='utf-8') as obj:
      obj.write("# cerc-hub export\n")
      obj.write(f'mtllib {mtl_name}\n')

      for building in self._city.buildings:
        obj.write(f'# building {building.name}\n')
        obj.write(f'g {building.name}\n')
        obj.write('s off\n')

        for surface in building.surfaces:
          obj.write(f'# surface {surface.name}\n')
          face = []
          normal = self._to_normal_vertex(surface.perimeter_polygon.coordinates)
          normal_index += 1
          textures = []
          for coordinate in surface.perimeter_polygon.coordinates:
            vertex = self._to_vertex(coordinate)
            if vertex not in vertices:
              vertex_index += 1
              vertices[vertex] = vertex_index
              current = vertex_index
              obj.write(vertex)
              textures.append(self._to_texture_vertex(coordinate))  # only append if non-existing
            else:
              current = vertices[vertex]
            face.append(f'{current}/{current}/{normal_index}')  # insert clockwise
          obj.writelines(normal)  # add the normal
          obj.writelines(textures)  # add the texture vertex

          faces.append(f"f {' '.join(face)}\n")
          obj.writelines(faces)
          faces = []



"""
Idf exports one building to idf format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Guillermo.GutierrezMorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
                   Oriol Gavalda Torrellas oriol.gavalda@concordia.ca
"""
import copy
import datetime
import glob
import os
from pathlib import Path
from geomeppy import IDF
import hub.helpers.constants as cte
from hub.city_model_structure.attributes.schedule import Schedule
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.helpers.configuration_helper import ConfigurationHelper


class Idf:
  """
  Exports city to IDF
  """
  _BUILDING = 'BUILDING'
  _ZONE = 'ZONE'
  _LIGHTS = 'LIGHTS'
  _APPLIANCES = 'OTHEREQUIPMENT'
  _PEOPLE = 'PEOPLE'
  _DHW = 'WATERUSE:EQUIPMENT'
  _THERMOSTAT = 'HVACTEMPLATE:THERMOSTAT'
  _IDEAL_LOAD_AIR_SYSTEM = 'HVACTEMPLATE:ZONE:IDEALLOADSAIRSYSTEM'
  _SURFACE = 'BUILDINGSURFACE:DETAILED'
  _SHADING = 'SHADING:BUILDING:DETAILED'
  _SHADING_PROPERTY = 'SHADINGPROPERTY:REFLECTANCE'
  _BUILDING_SURFACE = 'BuildingSurfaceDetailed'
  _CONSTRUCTION = 'CONSTRUCTION'
  _MATERIAL = 'MATERIAL'
  _MATERIAL_NOMASS = 'MATERIAL:NOMASS'
  _MATERIAL_ROOFVEGETATION = 'MATERIAL:ROOFVEGETATION'

  _WINDOW = 'FENESTRATIONSURFACE:DETAILED'
  _WINDOW_MATERIAL_SIMPLE = 'WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM'
  _ROUGHNESS = 'MediumRough'
  _INFILTRATION = 'ZONEINFILTRATION:DESIGNFLOWRATE'
  _VENTILATION = 'ZONEVENTILATION:DESIGNFLOWRATE'

  _HOURLY_SCHEDULE = 'SCHEDULE:DAY:HOURLY'
  _COMPACT_SCHEDULE = 'SCHEDULE:COMPACT'
  _FILE_SCHEDULE = 'SCHEDULE:FILE'
  _SCHEDULE_LIMIT = 'SCHEDULETYPELIMITS'
  _ON_OFF = 'On/Off'
  _FRACTION = 'Fraction'
  _ANY_NUMBER = 'Any Number'
  _CONTINUOUS = 'Continuous'
  _DISCRETE = 'Discrete'
  _SIZING_PERIODS = 'SIZINGPERIOD:DESIGNDAY'
  _LOCATION = 'SITE:LOCATION'
  _SIMPLE = 'Simple'

  idf_surfaces = {
    cte.WALL: 'wall',
    cte.GROUND: 'floor',
    cte.ROOF: 'roof'
  }
  idf_type_limits = {
    cte.ON_OFF: 'on/off',
    cte.FRACTION: 'Fraction',
    cte.ANY_NUMBER: 'Any Number',
    cte.CONTINUOUS: 'Continuous',
    cte.DISCRETE: 'Discrete'
  }
  idf_day_types = {
    cte.MONDAY: 'Monday',
    cte.TUESDAY: 'Tuesday',
    cte.WEDNESDAY: 'Wednesday',
    cte.THURSDAY: 'Thursday',
    cte.FRIDAY: 'Friday',
    cte.SATURDAY: 'Saturday',
    cte.SUNDAY: 'Sunday',
    cte.HOLIDAY: 'Holidays',
    cte.WINTER_DESIGN_DAY: 'WinterDesignDay',
    cte.SUMMER_DESIGN_DAY: 'SummerDesignDay'
  }

  def __init__(self, city, output_path, idf_file_path, idd_file_path, epw_file_path, export_type="Surfaces",
               target_buildings=None):
    self._city = city
    self._sanity_check()
    self._output_path = str(output_path.resolve())
    self._output_file = str((output_path / f'{city.name}.idf').resolve())
    self._export_type = export_type
    self._idd_file_path = str(idd_file_path)
    self._idf_file_path = str(idf_file_path)
    self._epw_file_path = str(epw_file_path)
    IDF.setiddname(self._idd_file_path)
    self._idf = IDF(self._idf_file_path, self._epw_file_path)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._ANY_NUMBER)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._FRACTION, Lower_Limit_Value=0.0, Upper_Limit_Value=1.0,
                           Numeric_Type=self._CONTINUOUS)
    self._idf.newidfobject(self._SCHEDULE_LIMIT, Name=self._ON_OFF, Lower_Limit_Value=0, Upper_Limit_Value=1,
                           Numeric_Type=self._DISCRETE)
    self._target_buildings = target_buildings
    self._adjacent_buildings = []
    if target_buildings is None:
      self._target_buildings = [building.name for building in self._city.buildings]
    else:
      for building_name in target_buildings:
        building = city.city_object(building_name)
        if building.neighbours is not None:
          self._adjacent_buildings += building.neighbours
    self._export()

  def _sanity_check(self):
    levels_of_detail = self._city.level_of_detail
    if levels_of_detail.geometry is None:
      raise AttributeError('Level of detail of geometry not assigned')
    if levels_of_detail.geometry < 1:
      raise AttributeError(f'Level of detail of geometry = {levels_of_detail.geometry}. Required minimum level 1')
    if levels_of_detail.construction is None:
      raise AttributeError('Level of detail of construction not assigned')
    if levels_of_detail.construction < 2:
      raise AttributeError(f'Level of detail of construction = {levels_of_detail.construction}. Required minimum level 2')
    if levels_of_detail.usage is None:
      raise AttributeError('Level of detail of usage not assigned')
    if levels_of_detail.usage < 2:
      raise AttributeError(f'Level of detail of usage = {levels_of_detail.usage}. Required minimum level 2')
    if levels_of_detail.weather is None:
      raise AttributeError('Level of detail of weather not assigned')
    if levels_of_detail.weather < 2:
      raise AttributeError(f'Level of detail of weather = {levels_of_detail.weather}. Required minimum level 2')

  @staticmethod
  def _matrix_to_list(points, lower_corner):
    lower_x = lower_corner[0]
    lower_y = lower_corner[1]
    lower_z = lower_corner[2]
    points_list = []
    for point in points:
      point_tuple = (point[0] - lower_x, point[1] - lower_y, point[2] - lower_z)
      points_list.append(point_tuple)
    return points_list

  @staticmethod
  def _matrix_to_2d_list(points):
    points_list = []
    for point in points:
      point_tuple = (point[0], point[1])
      points_list.append(point_tuple)
    return points_list

  def _add_material(self, layer):
    for material in self._idf.idfobjects[self._MATERIAL]:
      if material.Name == layer.material_name:
        return
    for material in self._idf.idfobjects[self._MATERIAL_NOMASS]:
      if material.Name == layer.material_name:
        return
    if layer.no_mass:
      self._idf.newidfobject(self._MATERIAL_NOMASS,
                             Name=layer.material_name,
                             Roughness=self._ROUGHNESS,
                             Thermal_Resistance=layer.thermal_resistance
                             )
    else:
      self._idf.newidfobject(self._MATERIAL,
                             Name=layer.material_name,
                             Roughness=self._ROUGHNESS,
                             Thickness=layer.thickness,
                             Conductivity=layer.conductivity,
                             Density=layer.density,
                             Specific_Heat=layer.specific_heat,
                             Thermal_Absorptance=layer.thermal_absorptance,
                             Solar_Absorptance=layer.solar_absorptance,
                             Visible_Absorptance=layer.visible_absorptance
                             )

  @staticmethod
  def _create_infiltration_schedules(thermal_zone):
    _infiltration_schedules = []
    if thermal_zone.thermal_control is None:
      return []
    for hvac_availability_schedule in thermal_zone.thermal_control.hvac_availability_schedules:
      _schedule = Schedule()
      _schedule.type = cte.INFILTRATION
      _schedule.data_type = cte.FRACTION
      _schedule.time_step = cte.HOUR
      _schedule.time_range = cte.DAY
      _schedule.day_types = copy.deepcopy(hvac_availability_schedule.day_types)
      _infiltration_values = []
      for hvac_value in hvac_availability_schedule.values:
        if hvac_value == 0:
          _infiltration_values.append(1.0)
        else:
          if thermal_zone.infiltration_rate_system_off == 0:
            _infiltration_values.append(0.0)
          else:
            _infiltration_values.append(
              thermal_zone.infiltration_rate_system_on / thermal_zone.infiltration_rate_system_off)
      _schedule.values = _infiltration_values
      _infiltration_schedules.append(_schedule)
    return _infiltration_schedules

  @staticmethod
  def _create_ventilation_schedules(thermal_zone):
    _ventilation_schedules = []
    if thermal_zone.thermal_control is None:
      return []
    for hvac_availability_schedule in thermal_zone.thermal_control.hvac_availability_schedules:
      _schedule = Schedule()
      _schedule.type = cte.VENTILATION
      _schedule.data_type = cte.FRACTION
      _schedule.time_step = cte.HOUR
      _schedule.time_range = cte.DAY
      _schedule.day_types = copy.deepcopy(hvac_availability_schedule.day_types)
      _ventilation_schedules = thermal_zone.thermal_control.hvac_availability_schedules
    return _ventilation_schedules

  @staticmethod
  def _create_yearly_values_schedules(schedule_type, values):
    _schedule = Schedule()
    _schedule.type = schedule_type
    _schedule.data_type = cte.ANY_NUMBER
    _schedule.time_step = cte.HOUR
    _schedule.time_range = cte.YEAR
    _schedule.day_types = ['monday',
                           'tuesday',
                           'wednesday',
                           'thursday',
                           'friday',
                           'saturday',
                           'sunday',
                           'holiday',
                           'winter_design_day',
                           'summer_design_day']
    _schedule.values = values
    return [_schedule]

  @staticmethod
  def _create_constant_value_schedules(schedule_type, value):
    _schedule = Schedule()
    _schedule.type = schedule_type
    _schedule.data_type = cte.ANY_NUMBER
    _schedule.time_step = cte.HOUR
    _schedule.time_range = cte.DAY
    _schedule.day_types = ['monday',
                           'tuesday',
                           'wednesday',
                           'thursday',
                           'friday',
                           'saturday',
                           'sunday',
                           'holiday',
                           'winter_design_day',
                           'summer_design_day']
    _schedule.values = [value for _ in range(0, 24)]
    return [_schedule]

  def _add_standard_compact_hourly_schedule(self, usage, schedule_type, schedules):
    for schedule in self._idf.idfobjects[self._COMPACT_SCHEDULE]:
      if schedule.Name == f'{schedule_type} schedules {usage}':
        return
    _kwargs = {'Name': f'{schedule_type} schedules {usage}',
               'Schedule_Type_Limits_Name': self.idf_type_limits[schedules[0].data_type],
               'Field_1': 'Through: 12/31'}
    counter = 1
    for j, schedule in enumerate(schedules):
      _val = schedule.values
      _new_field = ''
      for day_type in schedule.day_types:
        _new_field += f' {self.idf_day_types[day_type]}'
      _kwargs[f'Field_{j * 25 + 2}'] = f'For:{_new_field}'
      counter += 1
      for i, _ in enumerate(_val):
        _kwargs[f'Field_{j * 25 + 3 + i}'] = f'Until: {i + 1:02d}:00,{_val[i]}'
        counter += 1
    _kwargs[f'Field_{counter + 1}'] = 'For AllOtherDays'
    _kwargs[f'Field_{counter + 2}'] = 'Until: 24:00,0.0'
    self._idf.newidfobject(self._COMPACT_SCHEDULE, **_kwargs)

  def _write_schedules_file(self, schedule, usage):
    file_name = str((Path(self._output_path) / f'{schedule.type} schedules {usage.replace("/","_")}.csv').resolve())
    if not Path(file_name).exists():
      with open(file_name, 'w', encoding='utf8') as file:
        for value in schedule.values:
          file.write(f'{str(value)},\n')
    return Path(file_name).name

  def _add_file_schedule(self, usage, schedule, file_name):
    _schedule = self._idf.newidfobject(self._FILE_SCHEDULE, Name=f'{schedule.type} schedules {usage}')
    _schedule.Schedule_Type_Limits_Name = self.idf_type_limits[schedule.data_type]
    _schedule.File_Name = file_name
    _schedule.Column_Number = 1
    _schedule.Rows_to_Skip_at_Top = 0
    _schedule.Number_of_Hours_of_Data = 8760
    _schedule.Column_Separator = 'Comma'
    _schedule.Interpolate_to_Timestep = 'No'
    _schedule.Minutes_per_Item = 60

  def _add_schedules(self, usage, schedule_type, new_schedules):
    if len(new_schedules) < 1:
      return
    schedule_from_file = False
    for schedule in new_schedules:
      if len(schedule.values) > 168:  # Hours in one week
        schedule_from_file = True
    if schedule_from_file:
      for schedule in self._idf.idfobjects[self._FILE_SCHEDULE]:
        if schedule.Name == f'{schedule_type} schedules {usage}':
          return
      file_name = self._write_schedules_file(new_schedules[0], usage)
      self._add_file_schedule(usage, new_schedules[0], file_name)
      return

    for schedule in self._idf.idfobjects[self._HOURLY_SCHEDULE]:
      if schedule.Name == f'{schedule_type} schedules {usage}':
        return
    self._add_standard_compact_hourly_schedule(usage, schedule_type, new_schedules)
    return

  def _add_construction(self, thermal_boundary):
    for construction in self._idf.idfobjects[self._CONSTRUCTION]:
      if thermal_boundary.parent_surface.vegetation is not None:
        vegetation_name = f'{thermal_boundary.construction_name}_{thermal_boundary.parent_surface.vegetation.name}'
        if construction.Name == vegetation_name:
          return
      else:
        if construction.Name == f'{thermal_boundary.construction_name} {thermal_boundary.parent_surface.type}':
          return
    if thermal_boundary.layers is None:
      for material in self._idf.idfobjects[self._MATERIAL]:
        if material.Name == "DefaultMaterial":
          return

      self._idf.set_default_constructions()
      return
    for layer in thermal_boundary.layers:
      self._add_material(layer)
    layers = thermal_boundary.layers
    # The constructions should have at least one layer
    if thermal_boundary.parent_surface.vegetation is not None:
      vegetation_name = f'{thermal_boundary.construction_name}_{thermal_boundary.parent_surface.vegetation.name}'
      _kwargs = {'Name': vegetation_name,
                 'Outside_Layer': thermal_boundary.parent_surface.vegetation.name}
      for i in range(0, len(layers) - 1):
        _kwargs[f'Layer_{i + 2}'] = layers[i].material_name
    else:
      _kwargs = {'Name': f'{thermal_boundary.construction_name} {thermal_boundary.parent_surface.type}',
                 'Outside_Layer': layers[0].material_name}
      for i in range(1, len(layers) - 1):
        _kwargs[f'Layer_{i + 1}'] = layers[i].material_name
    self._idf.newidfobject(self._CONSTRUCTION, **_kwargs)

  def _add_window_construction_and_material(self, thermal_opening):
    for window_material in self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]:
      if window_material['UFactor'] == thermal_opening.overall_u_value and \
         window_material['Solar_Heat_Gain_Coefficient'] == thermal_opening.g_value:
        return

    order = str(len(self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]) + 1)
    material_name = 'glazing_' + order
    _kwargs = {'Name': material_name, 'UFactor': thermal_opening.overall_u_value,
               'Solar_Heat_Gain_Coefficient': thermal_opening.g_value}
    self._idf.newidfobject(self._WINDOW_MATERIAL_SIMPLE, **_kwargs)

    window_construction_name = 'window_construction_' + order
    _kwargs = {'Name': window_construction_name, 'Outside_Layer': material_name}
    self._idf.newidfobject(self._CONSTRUCTION, **_kwargs)

  def _add_zone(self, thermal_zone, name):
    for zone in self._idf.idfobjects['ZONE']:
      if zone.Name == name:
        return
    self._idf.newidfobject(self._ZONE, Name=name, Volume=thermal_zone.volume)
    self._add_heating_system(thermal_zone, name)

  def _add_thermostat(self, thermal_zone):
    thermostat_name = f'Thermostat {thermal_zone.usage_name}'
    for thermostat in self._idf.idfobjects[self._THERMOSTAT]:
      if thermostat.Name == thermostat_name:
        return thermostat
    return self._idf.newidfobject(
      self._THERMOSTAT,
      Name=thermostat_name,
      Heating_Setpoint_Schedule_Name=f'Heating thermostat schedules {thermal_zone.usage_name}',
      Cooling_Setpoint_Schedule_Name=f'Cooling thermostat schedules {thermal_zone.usage_name}'
    )

  def _add_heating_system(self, thermal_zone, zone_name):
    for air_system in self._idf.idfobjects[self._IDEAL_LOAD_AIR_SYSTEM]:
      if air_system.Zone_Name == zone_name:
        return
    thermostat = self._add_thermostat(thermal_zone)
    self._idf.newidfobject(self._IDEAL_LOAD_AIR_SYSTEM,
                           Zone_Name=zone_name,
                           System_Availability_Schedule_Name=f'Thermostat_availability schedules {thermal_zone.usage_name}',
                           Heating_Availability_Schedule_Name=f'Thermostat_availability schedules {thermal_zone.usage_name}',
                           Cooling_Availability_Schedule_Name=f'Thermostat_availability schedules {thermal_zone.usage_name}',
                           Template_Thermostat_Name=thermostat.Name)

  def _add_occupancy(self, thermal_zone, zone_name):
    number_of_people = thermal_zone.occupancy.occupancy_density * thermal_zone.total_floor_area
    fraction_radiant = 0
    total_sensible = (
      thermal_zone.occupancy.sensible_radiative_internal_gain + thermal_zone.occupancy.sensible_convective_internal_gain
    )
    if total_sensible != 0:
      fraction_radiant = thermal_zone.occupancy.sensible_radiative_internal_gain / total_sensible

    self._idf.newidfobject(self._PEOPLE,
                           Name=f'{zone_name}_occupancy',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Number_of_People_Schedule_Name=f'Occupancy schedules {thermal_zone.usage_name}',
                           Number_of_People_Calculation_Method="People",
                           Number_of_People=number_of_people,
                           Fraction_Radiant=fraction_radiant,
                           Activity_Level_Schedule_Name=f'Activity Level schedules {thermal_zone.usage_name}'
                           )

  def _add_lighting(self, thermal_zone: ThermalZone, zone_name: str):
    fraction_radiant = thermal_zone.lighting.radiative_fraction
    method = 'Watts/Area'
    storeys_number = int(thermal_zone.total_floor_area / thermal_zone.footprint_area)
    watts_per_zone_floor_area = thermal_zone.lighting.density * storeys_number
    subcategory = f'ELECTRIC EQUIPMENT#{zone_name}#GeneralLights'

    self._idf.newidfobject(self._LIGHTS,
                           Name=f'{zone_name}_lights',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Schedule_Name=f'Lighting schedules {thermal_zone.usage_name}',
                           Design_Level_Calculation_Method=method,
                           Watts_per_Zone_Floor_Area=watts_per_zone_floor_area,
                           Fraction_Radiant=fraction_radiant,
                           EndUse_Subcategory=subcategory
                           )

  def _add_appliances(self, thermal_zone, zone_name):
    fuel_type = 'Electricity'
    fraction_radiant = thermal_zone.appliances.radiative_fraction
    fraction_latent = thermal_zone.appliances.latent_fraction
    method = 'Watts/Area'
    storeys_number = int(thermal_zone.total_floor_area / thermal_zone.footprint_area)
    watts_per_zone_floor_area = thermal_zone.appliances.density * storeys_number
    subcategory = f'ELECTRIC EQUIPMENT#{zone_name}#InteriorEquipment'
    self._idf.newidfobject(self._APPLIANCES,
                           Fuel_Type=fuel_type,
                           Name=f'{zone_name}_appliance',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Schedule_Name=f'Appliance schedules {thermal_zone.usage_name}',
                           Design_Level_Calculation_Method=method,
                           Power_per_Zone_Floor_Area=watts_per_zone_floor_area,
                           Fraction_Latent=fraction_latent,
                           Fraction_Radiant=fraction_radiant,
                           EndUse_Subcategory=subcategory
                           )

  def _add_infiltration(self, thermal_zone, zone_name):
    schedule = f'INF_CONST schedules {thermal_zone.usage_name}'
    _infiltration = thermal_zone.infiltration_rate_system_off * cte.HOUR_TO_SECONDS
    self._idf.newidfobject(self._INFILTRATION,
                           Name=f'{zone_name}_infiltration',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Schedule_Name=schedule,
                           Design_Flow_Rate_Calculation_Method='AirChanges/Hour',
                           Air_Changes_per_Hour=_infiltration
                           )

  def _add_infiltration_surface(self, thermal_zone, zone_name):
    schedule = f'INF_CONST schedules {thermal_zone.usage_name}'
    _infiltration = thermal_zone.infiltration_rate_area_system_off* cte.INFILTRATION_75PA_TO_4PA
    self._idf.newidfobject(self._INFILTRATION,
                           Name=f'{zone_name}_infiltration',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Schedule_Name=schedule,
                           Design_Flow_Rate_Calculation_Method='Flow/ExteriorWallArea',
                           Flow_Rate_per_Exterior_Surface_Area=_infiltration
                           )

  def _add_ventilation(self, thermal_zone, zone_name):
    schedule = f'Ventilation schedules {thermal_zone.usage_name}'
    _air_change = thermal_zone.mechanical_air_change * cte.HOUR_TO_SECONDS
    self._idf.newidfobject(self._VENTILATION,
                           Name=f'{zone_name}_ventilation',
                           Zone_or_ZoneList_or_Space_or_SpaceList_Name=zone_name,
                           Schedule_Name=schedule,
                           Design_Flow_Rate_Calculation_Method='AirChanges/Hour',
                           Air_Changes_per_Hour=_air_change
                           )

  def _add_dhw(self, thermal_zone, zone_name, usage):
    peak_flow_rate = thermal_zone.domestic_hot_water.peak_flow * thermal_zone.total_floor_area
    self._idf.newidfobject(self._DHW,
                           Name=f'DHW {zone_name}',
                           Peak_Flow_Rate=peak_flow_rate,
                           Flow_Rate_Fraction_Schedule_Name=f'DHW_prof schedules {thermal_zone.usage_name}',
                           Target_Temperature_Schedule_Name=f'DHW_temp schedules {thermal_zone.usage_name}',
                           Hot_Water_Supply_Temperature_Schedule_Name=f'DHW_temp schedules {thermal_zone.usage_name}',
                           Cold_Water_Supply_Temperature_Schedule_Name=f'cold_temp schedules {usage}',
                           EndUse_Subcategory=f'DHW {zone_name}',
                           Zone_Name=zone_name
                           )

  def _rename_building(self, city_name):
    name = str(str(city_name.encode("utf-8")))
    for building in self._idf.idfobjects[self._BUILDING]:
      building.Name = f'Buildings in {name}'
      building['Solar_Distribution'] = 'FullExterior'

  def _remove_sizing_periods(self):
    while len(self._idf.idfobjects[self._SIZING_PERIODS]) > 0:
      self._idf.popidfobject(self._SIZING_PERIODS, 0)

  def _remove_location(self):
    self._idf.popidfobject(self._LOCATION, 0)

  def _export(self):
    """
    Export the idf file into the given path.

    If buildings to calculate are provided, only those will appear in the output variables, otherwise all the city
    buildings will be calculated.
    If adjacent buildings are provided those buildings will be calculated, but will not appear in the output variables.

    export type = "Surfaces|Block"
    """

    self._remove_location()
    self._remove_sizing_periods()
    self._rename_building(self._city.name)
    self._lod = self._city.level_of_detail.geometry
    for building in self._city.buildings:
      is_target = building.name in self._target_buildings or building.name in self._adjacent_buildings
      for internal_zone in building.internal_zones:
        if internal_zone.thermal_zones_from_internal_zones is None:
          self._target_buildings.remoidf_surface_typeve(building.name)
          is_target = False
          continue
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:

          for thermal_boundary in thermal_zone.thermal_boundaries:
            self._add_construction(thermal_boundary)
            if thermal_boundary.parent_surface.vegetation is not None:
              self._add_vegetation_material(thermal_boundary.parent_surface.vegetation)
            for thermal_opening in thermal_boundary.thermal_openings:
              self._add_window_construction_and_material(thermal_opening)

          if is_target:
            start = datetime.datetime.now()
            service_temperature = thermal_zone.domestic_hot_water.service_temperature
            usage = thermal_zone.usage_name
            _new_schedules = self._create_infiltration_schedules(thermal_zone)
            self._add_schedules(usage, 'Infiltration', _new_schedules)
            _new_schedules = self._create_ventilation_schedules(thermal_zone)
            self._add_schedules(usage, 'Ventilation', _new_schedules)
            self._add_schedules(usage, 'Occupancy', thermal_zone.occupancy.occupancy_schedules)
            self._add_schedules(usage, 'HVAC AVAIL', thermal_zone.thermal_control.hvac_availability_schedules)
            self._add_schedules(usage, 'Heating thermostat', thermal_zone.thermal_control.heating_set_point_schedules)
            self._add_schedules(usage, 'Cooling thermostat', thermal_zone.thermal_control.cooling_set_point_schedules)
            self._add_schedules(usage, 'Lighting', thermal_zone.lighting.schedules)
            self._add_schedules(usage, 'Appliance', thermal_zone.appliances.schedules)
            self._add_schedules(usage, 'DHW_prof', thermal_zone.domestic_hot_water.schedules)
            _new_schedules = self._create_yearly_values_schedules('cold_temp', building.cold_water_temperature[cte.HOUR])
            self._add_schedules(usage, 'cold_temp', _new_schedules)
            _new_schedules = self._create_constant_value_schedules('DHW_temp', service_temperature)
            self._add_schedules(usage, 'DHW_temp', _new_schedules)
            _new_schedules = self._create_constant_value_schedules('INF_CONST', 1)
            self._add_schedules(usage, 'INF_CONST', _new_schedules)
            _new_schedules = self._create_constant_value_schedules('Thermostat_availability', 1)
            self._add_schedules(usage, 'Thermostat_availability', _new_schedules)
            _occ = thermal_zone.occupancy
            if _occ.occupancy_density == 0:
              _total_heat = 0
            else:
              _total_heat = (_occ.sensible_convective_internal_gain + _occ.sensible_radiative_internal_gain
                             + _occ.latent_internal_gain) / _occ.occupancy_density
            _new_schedules = self._create_constant_value_schedules('Activity Level', _total_heat)
            self._add_schedules(usage, 'Activity Level', _new_schedules)
            self._add_zone(thermal_zone, building.name)
            self._add_heating_system(thermal_zone, building.name)
            self._add_infiltration_surface(thermal_zone, building.name)
            self._add_ventilation(thermal_zone, building.name)
            self._add_occupancy(thermal_zone, building.name)
            self._add_lighting(thermal_zone, building.name)
            self._add_appliances(thermal_zone, building.name)
            self._add_dhw(thermal_zone, building.name, usage)
      if self._export_type == "Surfaces":
        if is_target:
          if building.thermal_zones_from_internal_zones is not None:
            start = datetime.datetime.now()
            self._add_surfaces(building, building.name)
            print(f'add surfaces {datetime.datetime.now() - start}')
          else:
            self._add_pure_geometry(building, building.name)
        else:
          self._add_shading(building)
      else:
        self._add_block(building)

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Zone Ideal Loads Supply Air Total Heating Energy",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Zone Ideal Loads Supply Air Total Cooling Energy",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Water Use Equipment Heating Rate",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Zone Lights Electricity Rate",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Other Equipment Electricity Rate",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Zone Air Temperature",
      Reporting_Frequency="Hourly",
    )

    self._idf.newidfobject(
      "OUTPUT:VARIABLE",
      Variable_Name="Zone Air Relative Humidity",
      Reporting_Frequency="Hourly",
    )

    # post-process to erase windows associated to adiabatic walls
    windows_list = []
    for window in self._idf.idfobjects[self._WINDOW]:
      found = False
      for surface in self._idf.idfobjects[self._SURFACE]:
        if window.Building_Surface_Name == surface.Name:
          found = True
      if not found:
        windows_list.append(window)
    for window in windows_list:
      self._idf.removeidfobject(window)

    self._idf.saveas(str(self._output_file))
    return self._idf

  def run(self):
    """
    Start the energy plus simulation
    """
    self._idf.run(expandobjects=False, readvars=True, output_directory=self._output_path,
                  output_prefix=f'{self._city.name}_')

  def _add_block(self, building):
    _points = self._matrix_to_2d_list(building.foot_print.coordinates)
    self._idf.add_block(name=building.name, coordinates=_points, height=building.max_height,
                        num_stories=int(building.storeys_above_ground))

    for surface in self._idf.idfobjects[self._SURFACE]:
      for thermal_zone in building.thermal_zones_from_internal_zones:
        for boundary in thermal_zone.thermal_boundaries:
          if surface.Type == self.idf_surfaces[boundary.surface.type]:
            surface.Construction_Name = boundary.construction_name
            break
        for usage in thermal_zone.usages:
          surface.Zone_Name = usage.id
          break
        break
    self._idf.intersect_match()

  def _add_shading(self, building):
    for i, surface in enumerate(building.surfaces):
      shading = self._idf.newidfobject(self._SHADING, Name=f'{building.name}_{i}')
      coordinates = self._matrix_to_list(surface.solid_polygon.coordinates,
                                         self._city.lower_corner)
      shading.setcoords(coordinates)
      solar_reflectance = surface.short_wave_reflectance
      if solar_reflectance is None:
        solar_reflectance = ConfigurationHelper().short_wave_reflectance
      self._idf.newidfobject(self._SHADING_PROPERTY,
                             Shading_Surface_Name=f'{building.name}_{i}',
                             Diffuse_Solar_Reflectance_of_Unglazed_Part_of_Shading_Surface=solar_reflectance,
                             Fraction_of_Shading_Surface_That_Is_Glazed=0)

  def _add_pure_geometry(self, building, zone_name):
    for index, surface in enumerate(building.surfaces):
      outside_boundary_condition = 'Outdoors'
      sun_exposure = 'SunExposed'
      wind_exposure = 'WindExposed'
      idf_surface_type = self.idf_surfaces[surface.type]
      _kwargs = {'Name': f'Building_{building.name}_surface_{index}',
                 'Surface_Type': idf_surface_type,
                 'Zone_Name': zone_name}
      if surface.type == cte.GROUND:
        outside_boundary_condition = 'Ground'
        sun_exposure = 'NoSun'
        wind_exposure = 'NoWind'
      if surface.percentage_shared is not None and surface.percentage_shared > 0.5:
        outside_boundary_condition = 'Surface'
        outside_boundary_condition_object = f'Building_{building.name}_surface_{index}'
        sun_exposure = 'NoSun'
        wind_exposure = 'NoWind'
        _kwargs['Outside_Boundary_Condition_Object'] = outside_boundary_condition_object

      _kwargs['Outside_Boundary_Condition'] = outside_boundary_condition
      _kwargs['Sun_Exposure'] = sun_exposure
      _kwargs['Wind_Exposure'] = wind_exposure
      idf_surface = self._idf.newidfobject(self._SURFACE, **_kwargs)

      coordinates = self._matrix_to_list(surface.solid_polygon.coordinates,
                                         self._city.lower_corner)
      idf_surface.setcoords(coordinates)
    if self._lod >= 3:
      for internal_zone in building.internal_zones:
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          for boundary in thermal_zone.thermal_boundaries:
            self._add_windows_by_vertices(boundary)
    else:
      # idf only allows setting wwr for external walls
      wwr = 0
      try:
        self._idf.set_wwr(wwr, construction='window_construction_1')
      except ValueError:
        self._idf.set_wwr(0, construction='window_construction_1')

  def _add_surfaces(self, building, zone_name):
    for thermal_zone in building.thermal_zones_from_internal_zones:
      for index, boundary in enumerate(thermal_zone.thermal_boundaries):
        idf_surface_type = self.idf_surfaces[boundary.parent_surface.type]
        outside_boundary_condition = 'Outdoors'
        sun_exposure = 'SunExposed'
        wind_exposure = 'WindExposed'
        _kwargs = {'Name': f'Building_{building.name}_surface_{index}',
                   'Surface_Type': idf_surface_type,
                   'Zone_Name': zone_name}
        if boundary.parent_surface.type == cte.GROUND:
          outside_boundary_condition = 'Ground'
          sun_exposure = 'NoSun'
          wind_exposure = 'NoWind'
        if boundary.parent_surface.percentage_shared is not None and boundary.parent_surface.percentage_shared > 0.5:
          outside_boundary_condition = 'Surface'
          outside_boundary_condition_object = f'Building_{building.name}_surface_{index}'
          sun_exposure = 'NoSun'
          wind_exposure = 'NoWind'
          _kwargs['Outside_Boundary_Condition_Object'] = outside_boundary_condition_object
        _kwargs['Outside_Boundary_Condition'] = outside_boundary_condition
        _kwargs['Sun_Exposure'] = sun_exposure
        _kwargs['Wind_Exposure'] = wind_exposure

        if boundary.parent_surface.vegetation is not None:
          construction_name = f'{boundary.construction_name}_{boundary.parent_surface.vegetation.name}'
        else:
          construction_name = f'{boundary.construction_name} {boundary.parent_surface.type}'
        _kwargs['Construction_Name'] = construction_name

        surface = self._idf.newidfobject(self._SURFACE, **_kwargs)

        coordinates = self._matrix_to_list(boundary.parent_surface.solid_polygon.coordinates,
                                           self._city.lower_corner)
        surface.setcoords(coordinates)

    if self._lod >= 3:
      for internal_zone in building.internal_zones:
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          for boundary in thermal_zone.thermal_boundaries:
            self._add_windows_by_vertices(boundary)
    else:
      # idf only allows setting wwr for external walls
      wwr = 0
      for surface in building.surfaces:
        if surface.type == cte.WALL:
          wwr = surface.associated_thermal_boundaries[0].window_ratio
      self._idf.set_wwr(wwr, construction='window_construction_1')

  def _add_windows_by_vertices(self, boundary):
    raise NotImplementedError

  def _compare_window_constructions(self, window_construction, opening):
    glazing = window_construction['Outside_Layer']
    for material in self._idf.idfobjects[self._WINDOW_MATERIAL_SIMPLE]:
      if material['Name'] == glazing:
        if material['UFactor'] == opening.overall_u_value and material[
          'Solar_Heat_Gain_Coefficient'] == opening.g_value:
          return True
    return False

  def _add_vegetation_material(self, vegetation):
    for vegetation_material in self._idf.idfobjects[self._MATERIAL_ROOFVEGETATION]:
      if vegetation_material.Name == vegetation.name:
        return
    soil = vegetation.soil
    height = 0
    leaf_area_index = 0
    leaf_reflectivity = 0
    leaf_emissivity = 0
    minimal_stomatal_resistance = 0
    for plant in vegetation.plants:
      height += plant.percentage * plant.height
      leaf_area_index += plant.percentage * plant.leaf_area_index
      leaf_reflectivity += plant.percentage * plant.leaf_reflectivity
      leaf_emissivity += plant.percentage * plant.leaf_emissivity
      minimal_stomatal_resistance += plant.percentage * plant.minimal_stomatal_resistance
    self._idf.newidfobject(
      self._MATERIAL_ROOFVEGETATION,
      Name=vegetation.name,
      Height_of_Plants=height,
      Leaf_Area_Index=leaf_area_index,
      Leaf_Reflectivity=leaf_reflectivity,
      Leaf_Emissivity=leaf_emissivity,
      Minimum_Stomatal_Resistance=minimal_stomatal_resistance,
      Soil_Layer_Name=soil.name,
      Roughness=soil.roughness,
      Thickness=vegetation.soil_thickness,
      Conductivity_of_Dry_Soil=soil.dry_conductivity,
      Density_of_Dry_Soil=soil.dry_density,
      Specific_Heat_of_Dry_Soil=soil.dry_specific_heat,
      Thermal_Absorptance=soil.thermal_absorptance,
      Solar_Absorptance=soil.solar_absorptance,
      Visible_Absorptance=soil.visible_absorptance,
      Saturation_Volumetric_Moisture_Content_of_the_Soil_Layer=soil.saturation_volumetric_moisture_content,
      Residual_Volumetric_Moisture_Content_of_the_Soil_Layer=soil.residual_volumetric_moisture_content,
      Initial_Volumetric_Moisture_Content_of_the_Soil_Layer=soil.initial_volumetric_moisture_content,
      Moisture_Diffusion_Calculation_Method=self._SIMPLE
    )



"""
Thermal storage system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from hub.city_model_structure.energy_systems.energy_storage_system import EnergyStorageSystem
from hub.city_model_structure.building_demand.layer import Layer


class ThermalStorageSystem(EnergyStorageSystem):
  """"
  Thermal Storage System Class
  """

  def __init__(self):

    super().__init__()
    self._volume = None
    self._height = None
    self._layers = None
    self._maximum_operating_temperature = None
    self._heating_coil_capacity = None
    self._temperature = None
    self._heating_coil_energy_consumption = None

  @property
  def volume(self):
    """
    Get the physical volume of the storage system in cubic meters
    :return: float
    """
    return self._volume

  @volume.setter
  def volume(self, value):
    """
    Set the physical volume of the storage system in cubic meters
    :param value: float
    """
    self._volume = value

  @property
  def height(self):
    """
    Get the diameter of the storage system in meters
    :return: float
    """
    return self._height

  @height.setter
  def height(self, value):
    """
    Set the diameter of the storage system in meters
    :param value: float
    """
    self._height = value

  @property
  def layers(self) -> [Layer]:
    """
    Get construction layers
    :return: [layer]
    """
    return self._layers

  @layers.setter
  def layers(self, value):
    """
    Set construction layers
    :param value: [layer]
    """
    self._layers = value

  @property
  def maximum_operating_temperature(self):
    """
    Get maximum operating temperature of the storage system in degree Celsius
    :return: float
    """
    return self._maximum_operating_temperature

  @maximum_operating_temperature.setter
  def maximum_operating_temperature(self, value):
    """
    Set maximum operating temperature of the storage system in degree Celsius
    :param value: float
    """
    self._maximum_operating_temperature = value

  @property
  def heating_coil_capacity(self):
    """
    Get heating coil capacity in Watts
    :return: float
    """
    return self._heating_coil_capacity

  @heating_coil_capacity.setter
  def heating_coil_capacity(self, value):
    """
    Set heating coil capacity in Watts
    :param value: float
    """
    self._heating_coil_capacity = value

  @property
  def temperature(self) -> dict:
    """
    Get fuel consumption in W, m3, or kg
    :return: dict{[float]}
    """
    return self._temperature

  @temperature.setter
  def temperature(self, value):
    """
    Set fuel consumption in W, m3, or kg
    :param value: dict{[float]}
    """
    self._temperature = value

  @property
  def heating_coil_energy_consumption(self) -> dict:
    """
    Get fuel consumption in W, m3, or kg
    :return: dict{[float]}
    """
    return self._heating_coil_energy_consumption

  @heating_coil_energy_consumption.setter
  def heating_coil_energy_consumption(self, value):
    """
    Set fuel consumption in W, m3, or kg
    :param value: dict{[float]}
    """
    self._heating_coil_energy_consumption = value



"""
PV energy generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from hub.city_model_structure.energy_systems.generation_system import GenerationSystem


class PvGenerationSystem(GenerationSystem):
  """
  PvGenerationSystem class
  """
  def __init__(self):
    super().__init__()
    self._electricity_efficiency = None
    self._nominal_electricity_output = None
    self._nominal_ambient_temperature = None
    self._nominal_cell_temperature = None
    self._nominal_radiation = None
    self._standard_test_condition_cell_temperature = None
    self._standard_test_condition_maximum_power = None
    self._standard_test_condition_radiation = None
    self._cell_temperature_coefficient = None
    self._width = None
    self._height = None
    self._electricity_power_output = {}
    self._tilt_angle = None
    self._installed_capacity = None

  @property
  def nominal_electricity_output(self):
    """
    Get nominal_power_output of electricity generation devices or inverters in W
    :return: float
    """
    return self._nominal_electricity_output

  @nominal_electricity_output.setter
  def nominal_electricity_output(self, value):
    """
    Set nominal_power_output of electricity generation devices or inverters in W
    :param value: float
    """
    self._nominal_electricity_output = value

  @property
  def electricity_efficiency(self):
    """
    Get electricity_efficiency
    :return: float
    """
    return self._electricity_efficiency

  @electricity_efficiency.setter
  def electricity_efficiency(self, value):
    """
    Set electricity_efficiency
    :param value: float
    """
    self._electricity_efficiency = value

  @property
  def nominal_ambient_temperature(self):
    """
    Get nominal ambient temperature of PV panels in degree Celsius
    :return: float
    """
    return self._nominal_ambient_temperature

  @nominal_ambient_temperature.setter
  def nominal_ambient_temperature(self, value):
    """
    Set nominal ambient temperature of PV panels in degree Celsius
    :param value: float
    """
    self._nominal_ambient_temperature = value

  @property
  def nominal_cell_temperature(self):
    """
    Get nominal cell temperature of PV panels in degree Celsius
    :return: float
    """
    return self._nominal_cell_temperature

  @nominal_cell_temperature.setter
  def nominal_cell_temperature(self, value):
    """
    Set nominal cell temperature of PV panels in degree Celsius
    :param value: float
    """
    self._nominal_cell_temperature = value

  @property
  def nominal_radiation(self):
    """
    Get nominal radiation of PV panels
    :return: float
    """
    return self._nominal_radiation

  @nominal_radiation.setter
  def nominal_radiation(self, value):
    """
    Set nominal radiation of PV panels
    :param value: float
    """
    self._nominal_radiation = value

  @property
  def standard_test_condition_cell_temperature(self):
    """
    Get standard test condition cell temperature of PV panels in degree Celsius
    :return: float
    """
    return self._standard_test_condition_cell_temperature

  @standard_test_condition_cell_temperature.setter
  def standard_test_condition_cell_temperature(self, value):
    """
    Set standard test condition cell temperature of PV panels in degree Celsius
    :param value: float
    """
    self._standard_test_condition_cell_temperature = value

  @property
  def standard_test_condition_maximum_power(self):
    """
    Get standard test condition maximum power of PV panels in W
    :return: float
    """
    return self._standard_test_condition_maximum_power

  @standard_test_condition_maximum_power.setter
  def standard_test_condition_maximum_power(self, value):
    """
    Set standard test condition maximum power of PV panels in W
    :param value: float
    """
    self._standard_test_condition_maximum_power = value

  @property
  def standard_test_condition_radiation(self):
    """
    Get standard test condition radiation in W/m2
    :return: float
    """
    return self._standard_test_condition_radiation

  @standard_test_condition_radiation.setter
  def standard_test_condition_radiation(self, value):
    """
    Set standard test condition radiation in W/m2
    :param value: float
    """
    self._standard_test_condition_radiation = value

  @property
  def cell_temperature_coefficient(self):
    """
    Get cell temperature coefficient of PV module
    :return: float
    """
    return self._cell_temperature_coefficient

  @cell_temperature_coefficient.setter
  def cell_temperature_coefficient(self, value):
    """
    Set cell temperature coefficient of PV module
    :param value: float
    """
    self._cell_temperature_coefficient = value

  @property
  def width(self):
    """
    Get PV module width in m
    :return: float
    """
    return self._width

  @width.setter
  def width(self, value):
    """
    Set PV module width in m
    :param value: float
    """
    self._width = value

  @property
  def height(self):
    """
    Get PV module height in m
    :return: float
    """
    return self._height

  @height.setter
  def height(self, value):
    """
    Set PV module height in m
    :param value: float
    """
    self._height = value

  @property
  def electricity_power_output(self):
    """
    Get electricity_power in W
    :return: float
    """
    return self._electricity_power_output

  @electricity_power_output.setter
  def electricity_power_output(self, value):
    """
    Set electricity_power in W
    :param value: float
    """
    self._electricity_power_output = value

  @property
  def installed_capacity(self):
    """
    Get the total installed nominal capacity in W
    :return: float
    """
    return self._installed_capacity

  @installed_capacity.setter
  def installed_capacity(self, value):
    """
    Set the total installed nominal capacity in W
    :param value: float
    """
    self._installed_capacity = value



"""
Dictionaries module for montreal custom fuel to hub fuel
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import hub.helpers.constants as cte


class MontrealCustomFuelToHubFuel:
  """
  Montreal custom fuel to hub fuel class
  """

  def __init__(self):
    self._dictionary = {
      'gas': cte.GAS,
      'natural gas': cte.GAS,
      'biomass': cte.BIOMASS,
      'electricity': cte.ELECTRICITY,
      'renewable': cte.RENEWABLE,
      'butane': cte.BUTANE,
      'diesel': cte.DIESEL
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary
